import importlib
from glob import glob
import logging
import os
import sys
import tempfile
import subprocess

import yaml
from aiohttp_session import get_session
from aiohttp import web

from brome.core.settings import BROME_CONFIG
from brome.core import exceptions
from brome.webserver.server.server_decorator import (
    require,
    exception_handler
)

logger = logging.getLogger('bromewebserver')


def get_tests():
    test_list = []
    script_test_prefix = BROME_CONFIG['brome']['script_test_prefix']

    tests_dir = os.path.join(
        BROME_CONFIG['project']['absolute_path'],
        BROME_CONFIG['brome']['script_folder_name']
    )

    if os.path.isdir(tests_dir):
        tests = glob(
            os.path.join(tests_dir, '%s*.py' % script_test_prefix)
        )
        for test in sorted(tests):
            name = test.split(os.sep)[-1][len(script_test_prefix):-3]
            test_list.append(name)

    return test_list


class LogStreamOut(web.View):

    def import_model(self, model):
        try:
            m = importlib.import_module(
                'brome.model.{model}'.format(model=model)
            )
            return getattr(m, model.title())
        except ImportError:
            raise exceptions.ModelImportException(
                '{model} not found'.format(model=model)
            )

    @exception_handler()
    @require('login')
    async def post(self):
        essential_keys = [
            'skip',
            'model',
            'uid'
        ]

        # REQUEST
        req_data = await self.request.json()

        # Keys validation
        for key in essential_keys:
            if key not in req_data.keys():
                raise exceptions.InvalidRequestException(
                    "Missing '{key}' params".format(
                        key=key
                    )
                )

        skip = req_data.get('skip')
        model_name = req_data.get('model')
        uid = req_data.get('uid')

        if model_name not in ['testbatch', 'testinstance']:
            raise exceptions.InvalidRequestException(
                "Only model 'testbatch' and 'testinstance' are allowed"
            )

        model_class = self.import_model(model_name)
        query = self.request.db_session.query(model_class)\
            .filter(model_class.mongo_id == uid)

        if not query.count():
            raise exceptions.InvalidRequestException(
                "Instance not found for uid: {uid}".format(
                    uid=uid
                )
            )

        parent = query.one()

        log_file_path = parent.log_file_path
        if not log_file_path:
            raise exceptions.InvalidRequestException(
                "The instance 'log_file_path' is empty"
            )

        full_log_file_path = os.path.join(
            parent.root_path,
            log_file_path
        )

        results = []
        total = 0
        with open(full_log_file_path) as fd:
            for _ in range(skip):
                total += 1
                next(fd)

            results = fd.readlines()
            total += len(results)

        name = log_file_path\
            .rsplit(os.sep)[-1]\
            .rsplit('.')[0]\
            .replace('_', ' ')

        context = {
            'db_session': self.request.db_session,
            'ws_session': await get_session(self.request),
            'method': 'read'
        }

        # RESPONSE
        response_data = {
            'success': True,
            'total': total,
            'name': name,
            'parent': await parent.serialize(context),
            'results': results
        }
        return web.json_response(response_data)


class GetBromeConfig(web.View):

    @exception_handler()
    @require('login')
    async def get(self):
        # RESPONSE
        response_data = {
            'success': True,
            'config': BROME_CONFIG,
            'tests': get_tests()
        }
        return web.json_response(response_data)


class StartTestBatch(web.View):

    @exception_handler()
    @require('login')
    async def post(self):
        req_data = await self.request.json()

        tests = req_data.get('tests', get_tests())
        browsers = req_data['browsers']

        runner_path = os.path.join(
            BROME_CONFIG['project']['absolute_path'],
            BROME_CONFIG["brome"]["brome_executable_name"]
        )

        test_file_path = os.path.join(
            tempfile.gettempdir(),
            'test_file.yaml'
        )
        with open(test_file_path, 'w') as f:
            f.write(yaml.dump(tests, default_flow_style=False))

        commands = [
            sys.executable,
            runner_path,
            "run",
            "-r",
            ','.join(browsers),
            "--test-file",
            test_file_path
        ]
        subprocess.Popen(
            commands,
            stdout=open('runner.log', 'a'),
            stderr=open('runner.log', 'a')
        )
        # TODO return test batch id
        # RESPONSE
        response_data = {
            'success': True
        }
        return web.json_response(response_data)



from aiohttp_session import get_session

from brome.model.user import User


def permits(request, session, permission):
    # GET THE USER
    uid = session.get('uid')
    if not uid:
        return False

    user_query = request.db_session.query(User).filter(User.mongo_id == uid)
    if user_query.count():
        user = user_query.one()
        if user.enable:
            # LOGIN
            if permission == 'login':
                return True

            # ADMIN
            elif permission == 'admin':
                if user.role == 'admin':
                    return True

    # BY DEFAUlT
    return False


def get_user_from_session(session, db_session):
    try:
        return db_session.query(User)\
            .filter(User.mongo_id == session['uid']).one()
    except:
        return None


async def set_session(user, request):
    session = await get_session(request)
    session['uid'] = user.get_uid()



from brome.core.exceptions import BromeBrowserConfigException


class BrowserConfig(object):
    """Class that hold the effective browser config

    Attributes:
        runner (object)
        browser_id (str)
        browsers_config (dict)
    """
    def __init__(self, runner, browser_id, browsers_config):
        self.runner = runner
        self.browser_id = browser_id
        self.browsers_config = browsers_config

        # Get the effective browser config
        self.config = self.browsers_config[self.browser_id]

        # LOCATION
        if self.config.get("amiid"):
            self.location = 'ec2'

        elif self.config.get("vbname"):
            self.location = 'virtualbox'

        elif self.config.get('appium'):
            self.location = 'appium'

        elif self.config.get('saucelabs'):
            self.location = 'saucelabs'

        elif self.config.get('browserstack'):
            self.location = 'browserstack'

        else:
            self.location = 'localhost'

        # Validation
        self.validate_config()

    def validate_config(self):
        """Validate that the browser config contains all the needed config
        """

        # LOCALHOST
        if self.location == 'localhost':
            if 'browserName' not in self.config.keys():
                msg = "Add the 'browserName' in your local_config: e.g.: 'Firefox', 'Chrome', 'Safari'"  # noqa
                self.runner.critical_log(msg)
                raise BromeBrowserConfigException(msg)

        # EC2
        elif self.location == 'ec2':
            self.validate_ec2_browser_config()

        # VIRTUALBOX
        elif self.location == 'virtualbox':
            self.validate_virtualbox_config()

    def get_id(self):
        return self.browser_id

    def get(self, key, *args):
        return self.config.get(key, *args)

    def get_platform(self):
        if 'platform' in self.config:
            return self.config['platform']

        elif 'platformName' in self.config:
            return self.config['platformName']

        else:
            msg = "Platform could not be determined from the browser config"
            self.runner.error_log(msg)
            raise BromeBrowserConfigException(msg)

    def validate_ec2_browser_config(self):
        """Validate that the ec2 config is conform
        """

        if self.config.get('launch', True):
            required_keys = [
                'browserName',
                'platform',
                'ssh_key_path',
                'username',
                'amiid',
                'region',
                'instance_type',
                'security_group_ids',
                'selenium_command'
            ]
        else:
            required_keys = [
                'browserName',
                'platform'
            ]

        for key in required_keys:
            if key not in self.config.keys():
                msg = "Add the config: %s" % key
                self.runner.critical_log(msg)
                raise BromeBrowserConfigException(msg)

        optional_keys = {
            'terminate': True,
            'launch': True,
            'record_session': False,
            'vnc_port': 5900,
            'vnc_password': '',
            'vnc_password_file': '~/.vnc/passwd',
            'nb_browser_by_instance': 1,
            'max_number_of_instance': 1,
            'hub_ip': 'localhost'
        }
        for key, default in iter(optional_keys.items()):
            if key not in self.config.keys():
                self.runner.warning_log(
                    "Missing config: %s; using default: %s" %
                    (key, default)
                )
                self.config[key] = default

    def validate_virtualbox_config(self):
        """Validate that the virtualbox config is conform
        """
        if self.config.get('launch', True):
            required_keys = [
                'browserName',
                'platform',
                'username',
                'password',
                'selenium_command'
            ]
        else:
            required_keys = [
                'browserName',
                'platform'
            ]

        for key in required_keys:
            if key not in self.config.keys():
                msg = "Add the config: %s" % key
                self.runner.critical_log(msg)
                raise BromeBrowserConfigException(msg)

        optional_keys = {
            'terminate': True,
            'launch': True,
            'record_session': False,
            'vbox_type': 'gui',
            'vnc_port': 5900,
            'vnc_password': '',
            'vnc_password_file': '~/.vnc/passwd',
            'nb_browser_by_instance': 1,
            'max_number_of_instance': 1,
            'hub_ip': 'localhost'
        }
        for key, default in iter(optional_keys.items()):
            if key not in self.config.keys():
                self.runner.warning_log(
                    "Missing config: %s; using default: %s" % (key, default)
                )
                self.config[key] = default



#! /usr/bin/env python

import os
import asyncio
from logging.handlers import TimedRotatingFileHandler
import logging

from redis import Redis
from rq import Queue
import aioredis
from aiohttp_session import redis_storage, session_middleware
import jinja2
import aiohttp_jinja2
from aiohttp import web

from brome.webserver.server.routes import routes
from brome.webserver.server.middlewares import db_handler
from brome.core.settings import BROME_CONFIG

logger = logging.getLogger('bromewebserver')

async def shutdown(server, app, handler):

    server.close()
    await server.wait_closed()
    await app.shutdown()
    await handler.finish_connections(10.0)
    await app.cleanup()
    await app.redis_pool.clear()


async def init(loop):
    ROOT = os.path.join(
        os.path.abspath(os.path.dirname(__file__)),
        '..'
    )

    # CONFIG
    logger = logging.getLogger('bromewebserver')
    logger.debug('Env: {env}'.format(env=BROME_CONFIG['webserver']['env']))

    # LOGGER
    logger.setLevel(
        getattr(logging, BROME_CONFIG['webserver'].get('log_level', 'INFO'))
    )

    formatter = logging.Formatter(
        '[L:%(lineno)d]# %(levelname)-8s [%(asctime)s]  %(message)s',
        datefmt='%d-%m-%Y %H:%M:%S'
    )

    # StreamHandler
    if BROME_CONFIG['webserver'].get('streamlogger', True):
        sh = logging.StreamHandler()
        sh.setFormatter(formatter)
        logger.addHandler(sh)

    # FileHandler
    if BROME_CONFIG['webserver'].get('filelogger', False):
        if BROME_CONFIG['webserver'].get('log_file_path'):
            fh = TimedRotatingFileHandler(
                BROME_CONFIG['webserver'].get('log_file_path'),
                when="midnight"
            )
            fh.setFormatter(formatter)
            logger.addHandler(fh)

    # SESSION
    redis_pool = await aioredis.create_pool(('localhost', 6379))
    storage = redis_storage.RedisStorage(redis_pool)
    app = web.Application(loop=loop, middlewares=[
        session_middleware(storage),
        db_handler
    ])

    app.redis_pool = redis_pool

    # QUEUE
    app.queue = Queue(connection=Redis())

    handler = app.make_handler()

    # ROUTES
    for route in routes:
        app.router.add_route(route[0], route[1], route[2], name=route[3])

    if BROME_CONFIG['webserver'].get('env', 'production') in \
            ['development']:
        static_path = os.path.join(ROOT, 'dist-dev')
        try:
            if BROME_CONFIG['project']['test_batch_result_path']:
                os.symlink(
                    os.path.join(
                        BROME_CONFIG['project']['test_batch_result_path'],
                        'tb_results'
                    ),
                    os.path.join(static_path, 'static', 'tb_results')
                )
                if BROME_CONFIG['webserver'].get('env') != 'test':
                    logger.info(
                        "Serving test batch results from symlink: %s"
                        % BROME_CONFIG['project']['test_batch_result_path']
                    )
        except FileExistsError:
            pass

    elif BROME_CONFIG['webserver']['env'] == 'test':
        # NOTE don't serve any static in test mode
        static_path = False
    else:
        if BROME_CONFIG['webserver'].get('release', 'latest') == 'latest':
            latest_version_path = os.path.join(
                ROOT,
                'releases',
                'latest.txt'
            )
            if os.path.exists(latest_version_path):
                with open(latest_version_path, 'r') as fd:
                    release_version = fd.read()
            else:
                raise Exception("The latest.txt file doesn't exists")
        else:
            release_version = BROME_CONFIG['webserver'].get('release')

        static_path = os.path.join(ROOT, 'releases', release_version)

    if static_path:
        app.router.add_static('/', static_path, name='static')

        if BROME_CONFIG['webserver'].get('env') != 'test':
            logger.info(
                "Serving static: {static_path}"
                .format(
                    static_path=static_path
                )
            )

        aiohttp_jinja2.setup(app, loader=jinja2.FileSystemLoader(static_path))

    # PREPARE HOOK
    async def after_request(request, response):
        if hasattr(request, 'db_session'):
            request.db_session.end()
            # request.db_session.db.connection.disconnect()

    app.on_response_prepare.append(after_request)

    serv_generator = loop.create_server(
        handler,
        BROME_CONFIG['webserver'].get('host'),
        BROME_CONFIG['webserver'].get('port')
    )

    return serv_generator, handler, app


def run_app():
    loop = asyncio.get_event_loop()

    serv_generator, handler, app = loop.run_until_complete(
        init(loop)
    )

    serv = loop.run_until_complete(serv_generator)

    logger.info('Server listening at %s' % str(serv.sockets[0].getsockname()))
    try:
        loop.run_forever()

    except KeyboardInterrupt:
        logger.debug('Server stopping...')

    finally:
        loop.run_until_complete(shutdown(serv, app, handler))
        loop.close()

    logger.debug('Server stopped')

if __name__ == '__main__':
    run_app()



import sys
import os
import re

def search(pattern, string):
    try:
        re.search(pattern, string).group(1)
        return True
    except (IndexError, AttributeError):
        return False

def print_result(print_header, header, print_lineno, lineno, print_line, line, return_string):
    result = ''
    if print_header:
        result += '%s' % header
    if print_lineno:
        if len(result) > 0:
            result += ':'
        result += '%d' % lineno
    if print_line:
        if len(result) > 0:
            result += ':'
        result += line

    if return_string:
        return result.strip('\n')
    else:
        sys.stdout.write('%s\n' % result.strip('\n'))

def grep_file(filename, pattern, print_headers, print_lineno, print_lines, return_list):
    if return_list:
        result_list = []

    with open(filename, 'r') as fd:
        for i, line in enumerate(fd.readlines()):
            found = search(pattern, line)
            if found:
                if return_list:
                    result_list.append(print_result(print_headers, filename, print_lineno, i, print_lines, line, True))
                else:
                    print_result(print_headers, filename, print_lineno, i, print_lines, line, False)

                if print_headers and not print_lines: # files-with-matches option
                    break

    if return_list:
        return result_list

def grep_files(paths, pattern, recursive, print_headers=True, print_lineno=True, print_line=True, return_list = False):
    if return_list:
        result_list = []

    for path in paths:
        if os.path.isfile(path):
            if path.endswith(".py"):
                result = grep_file(path, pattern, print_headers, print_lineno, print_line, return_list)
                if return_list:
                    result_list.extend(result)

        elif os.path.isdir(path):
            if recursive:
                more_paths = [os.path.join(path, child) for child in os.listdir(path)]
                result = grep_files(more_paths, pattern, recursive, print_headers, print_lineno, print_line, return_list)
                if return_list:
                    result_list.extend(result)
            else:
                sys.stdout.write('grep: %s: Is a directory\n' % file_)

    if return_list:
        return result_list



class Stateful(object):
    def __getstate__(self):
        dict_ = {}

        for key, value in iter(self.__dict__.items()):
            if type(value) in [str, int, float, dict, list]:
                dict_[key] = value

        return dict_

    @classmethod
    def is_stateful(cls, value):
        if Stateful in value.__class__.__bases__ or \
                type(value) in [int, str, dict, list, float]:
            return True
        else:
            return False

    @classmethod
    def cleanup_list(cls, list_):
        effective_list = []
        for item in list_:
            if type(item) == list:
                effective_list.append(cls.cleanup_list(item))
            elif type(item) == dict:
                effective_list.append(cls.cleanup_dict(item))
            else:
                ret = cls.is_stateful(item)
                if ret:
                    effective_list.append(item)

        return effective_list

    @classmethod
    def cleanup_dict(cls, dict_):
        effective_dict = {}
        for key, value in iter(dict_.items()):
            if type(value) == list:
                effective_dict[key] = cls.cleanup_list(value)
            elif type(value) == dict:
                effective_dict[key] = cls.cleanup_dict(value)
            else:
                ret = cls.is_stateful(value)
                if ret:
                    effective_dict[key] = value

        return effective_dict

    @classmethod
    def cleanup_state(cls, state):
        effective_state = {}
        for key, value in iter(state.items()):
            if type(value) == list:
                effective_state[key] = cls.cleanup_list(value)
            elif type(value) == dict:
                effective_state[key] = cls.cleanup_dict(value)
            else:
                ret = cls.is_stateful(value)
                if ret:
                    effective_state[key] = value

        return effective_state



from mongoalchemy.fields import (
    StringField,
    BoolField,
    AnythingField,
    ObjectIdField,
    DictField
)


from brome.model.basemodel import BaseModel
from brome.model.testinstance import Testinstance
from brome.core import exceptions


class Testresult(BaseModel):
    result = BoolField()
    browser_capabilities = DictField(AnythingField())
    browser_id = StringField()
    title = StringField()
    testid = StringField(default='')

    root_path = StringField(default='')
    screenshot_path = StringField(default='')
    video_capture_path = StringField(default='')

    extra_data = DictField(StringField(), default=dict())

    test_id = ObjectIdField(required=False)
    test_instance_id = ObjectIdField()
    test_batch_id = ObjectIdField()

    def __repr__(self):
        try:
            return "Testresult <uid: {self.uid}><result: {self.result}><browser_id: {self.browser_id}><title: {self.title}>".format(  # noqa
                self=self
            )
        except AttributeError:
            return "Testresult uninitialized"

    async def sanitize_data(self, context):
        return []

    def get_video_current_time(self, context):
        db_session = context.get('db_session')

        test_instance = db_session.query(Testinstance)\
            .filter(Testinstance.mongo_id == self.test_instance_id)\
            .one()

        created_ts = self.created_ts
        test_instance_created_ts = test_instance.created_ts

        current_time = (created_ts - test_instance_created_ts).total_seconds()

        return current_time - 2

    async def serialize(self, context):
        data = {}
        data['uid'] = self.get_uid()
        data['result'] = self.result
        data['browser_capabilities'] = self.browser_capabilities
        data['root_path'] = self.root_path
        data['screenshot_path'] = self.screenshot_path
        data['video_capture_path'] = self.video_capture_path
        data['video_capture_current_time'] = self.get_video_current_time(
            context
        )
        data['extra_data'] = self.extra_data
        data['title'] = self.title
        data['testid'] = self.testid
        data['test_instance_id'] = str(self.test_instance_id)
        data['test_batch_id'] = str(self.test_batch_id)
        return data

    async def method_autorized(self, context):
        method = context.get('method')
        if method in ['create', 'update', 'delete']:
            return False
        else:
            return True

    async def validate_and_save(self, context):
        data = context.get('data')
        db_session = context.get('db_session')

        is_new = await self.is_new()

        # RESULT
        result = data.get('result')
        if result is not None:
            self.result = result
        else:
            if is_new:
                raise exceptions.MissingModelValueException('result')

        # BROWSER ID
        browser_id = data.get('browser_id')
        if browser_id is not None:
            self.browser_id = browser_id
        else:
            if is_new:
                raise exceptions.MissingModelValueException('browser_id')

        # BROWSER CAPABILITIES
        browser_capabilities = data.get('browser_capabilities')
        if browser_capabilities is not None:
            self.browser_capabilities = browser_capabilities
        else:
            if is_new:
                raise exceptions.MissingModelValueException(
                    'browser_capabilities'
                )

        # ROOT PATH
        root_path = data.get('root_path')
        if root_path is not None:
            self.root_path = root_path

        # SCREENSHOT PATH
        screenshot_path = data.get('screenshot_path')
        if screenshot_path is not None:
            self.screenshot_path = screenshot_path

        # VIDEOCAPTURE PATH
        video_capture_path = data.get('video_capture_path')
        if video_capture_path is not None:
            self.video_capture_path = video_capture_path

        # EXTRA DATA
        extra_data = data.get('extra_data')
        if extra_data is not None:
            self.extra_data = extra_data

        # TITLE
        title = data.get('title')
        if title is not None:
            self.title = title
        else:
            if is_new:
                raise exceptions.MissingModelValueException('title')

        # TESTID
        testid = data.get('testid')
        if testid is not None:
            self.testid = testid
        else:
            if is_new:
                raise exceptions.MissingModelValueException('testid')

        # TEST ID
        test_id = data.get('test_id')
        if test_id is not None:
            self.test_id = test_id

        # TEST INSTANCE ID
        test_instance_id = data.get('test_instance_id')
        if test_instance_id is not None:
            self.test_instance_id = test_instance_id
        else:
            if is_new:
                raise exceptions.MissingModelValueException('test_instance_id')

        # TEST BATCH ID
        test_batch_id = data.get('test_batch_id')
        if test_batch_id is not None:
            self.test_batch_id = test_batch_id
        else:
            if is_new:
                raise exceptions.MissingModelValueException('test_batch_id')

        db_session.save(self, safe=True)



from datetime import datetime

from mongoalchemy.fields import (
    DateTimeField,
    BoolField,
    AnythingField,
    StringField,
    ObjectIdField,
    DictField
)


from brome.model.basemodel import BaseModel
from brome.core.utils import convert_tz_datetime
from brome.core import exceptions


class Testinstance(BaseModel):

    name = StringField()
    browser_capabilities = DictField(AnythingField())
    browser_id = StringField()
    starting_timestamp = DateTimeField(use_tz=True)
    ending_timestamp = DateTimeField(required=False, use_tz=True)
    terminated = BoolField(default=False)
    extra_data = DictField(StringField(), default=dict())

    root_path = StringField(default='')
    log_file_path = StringField(default='')
    network_capture_path = StringField(default='')
    video_capture_path = StringField(default='')
    # Bot diary

    test_batch_id = ObjectIdField()

    def __repr__(self):
        try:
            return "Testinstance <uid: {self.mongo_id}><test_batch_id: {self.test_batch_id}>".format(  # noqa
                self=self
            )
        except AttributeError:
            return "Testinstance uninitialized"

    async def sanitize_data(self, context):
        return []

    async def serialize(self, context):
        ws_session = context.get('ws_session')

        data = {}
        data['uid'] = self.get_uid()
        data['name'] = self.name
        data['root_path'] = self.root_path
        data['log_file_path'] = self.log_file_path
        data['network_capture_path'] = self.network_capture_path
        data['video_capture_path'] = self.video_capture_path
        data['browser_capabilities'] = self.browser_capabilities
        data['starting_timestamp'] = convert_tz_datetime(
            self.starting_timestamp,
            ws_session.get('tz')
        ).isoformat()
        data['terminated'] = self.terminated
        if self.terminated:
            data['ending_timestamp'] = convert_tz_datetime(
                self.ending_timestamp,
                ws_session.get('tz')
            ).isoformat()
        else:
            data['ending_timestamp'] = False
        data['extra_data'] = self.extra_data
        data['test_batch_id'] = str(self.test_batch_id)
        return data

    async def method_autorized(self, context):
        method = context.get('method')
        if method in ['create', 'update', 'delete']:
            return False
        else:
            return True

    async def validate_and_save(self, context):
        data = context.get('data')
        db_session = context.get('db_session')

        is_new = await self.is_new()

        # NAME
        name = data.get('name')
        if name:
            self.name = name
        else:
            if is_new:
                raise exceptions.MissingModelValueException('name')

        # BROWSER ID
        browser_id = data.get('browser_id')
        if browser_id:
            self.browser_id = browser_id
        else:
            if is_new:
                raise exceptions.MissingModelValueException('browser_id')

        # BROWSER CAPABILITIES
        browser_capabilities = data.get('browser_capabilities')
        if browser_capabilities:
            self.browser_capabilities = browser_capabilities
        else:
            if is_new:
                raise exceptions.MissingModelValueException(
                    'browser_capabilities'
                )

        # STARTING TIMESTAMP
        starting_timestamp = data.get('starting_timestamp')
        if starting_timestamp:
            self.starting_timestamp = starting_timestamp
        else:
            if is_new:
                self.starting_timestamp = datetime.now()

        # ENDING TIMESTAMP
        ending_timestamp = data.get('ending_timestamp')
        if ending_timestamp:
            self.ending_timestamp = ending_timestamp

        # EXTRA DATA
        extra_data = data.get('extra_data')
        if extra_data is not None:
            self.extra_data = extra_data

        # ROOT PATH
        root_path = data.get('root_path')
        if root_path is not None:
            self.root_path = root_path

        # LOG FILE PATH
        log_file_path = data.get('log_file_path')
        if log_file_path is not None:
            self.log_file_path = log_file_path

        # NETWORk CAPTURE DATA
        network_capture_data = data.get('network_capture_data')
        if network_capture_data is not None:
            self.network_capture_data = network_capture_data

        # VIDEO CAPTURE PATH
        video_capture_path = data.get('video_capture_path')
        if video_capture_path is not None:
            self.video_capture_path = video_capture_path

        # TEST BATCH ID
        test_batch_id = data.get('test_batch_id')
        if test_batch_id:
            self.test_batch_id = test_batch_id
        else:
            if is_new:
                raise exceptions.MissingModelValueException('test_batch_id')

        db_session.save(self, safe=True)



import logging

from aiohttp_session import get_session
from aiohttp import web

from brome.core import exceptions
from brome.model.user import User
from brome.core.settings import BROME_CONFIG
from brome.webserver.server.server_decorator import (
    require,
    exception_handler,
    csrf_protected
)
from brome.webserver.server.auth import set_session, get_user_from_session

logger = logging.getLogger('bromewebserver')


class Login(web.View):

    @exception_handler()
    @csrf_protected()
    async def post(self):
        try:
            data = await self.request.json()
            email = data['email']
            password = data['password']
        except:
            raise exceptions.InvalidRequestException('No json send')

        query = self.request.db_session.query(User)\
            .filter(User.email == email)
        if query.count():
            user = query.one()
            is_password_valid = await user.check_password(password)
            is_enable = user.enable
            if is_password_valid and is_enable:
                await set_session(user, self.request)
                session = await get_session(self.request)
                session['tz'] = data.get('user_timezone')

                context = {
                    'db_session': self.request.db_session,
                    'ws_session': session,
                    'method': 'read',
                    'queue': self.request.app.queue
                }

                resp_data = {
                    'success': True,
                    'token': session['csrf_token'],
                    'user': await user.serialize(context)
                }
            else:
                raise exceptions.WrongEmailOrPasswordException()
        else:
            raise exceptions.WrongEmailOrPasswordException(
                "Wrong email: '{email}'".format(email=email)
            )

        return web.json_response(resp_data)


class Register(web.View):

    @exception_handler()
    @csrf_protected()
    async def post(self):
        try:
            data = await self.request.json()
        except:
            raise exceptions.InvalidRequestException('No json send')

        context = {
            'db_session': self.request.db_session,
            'method': 'create',
            'queue': self.request.app.queue
        }

        registration_token = data.get('registration_token')
        if registration_token != \
                BROME_CONFIG['webserver']['registration_token']:
            raise exceptions.InvalidRegistrationTokenException()

        # INIT USER
        user = User()
        context['data'] = data
        sane_data = await user.sanitize_data(context)
        context['data'] = sane_data
        await user.validate_and_save(context)

        # SET SESSION
        await set_session(user, self.request)
        session = await get_session(self.request)
        session['tz'] = data.get('user_timezone')

        context['method'] = 'read'
        context['user'] = user
        resp_data = {
            'success': True,
            'user': await user.serialize(context),
            'token': session['csrf_token']
        }
        return web.json_response(resp_data)


class Logout(web.View):

    @exception_handler()
    @require('login')
    @csrf_protected()
    async def post(self):
        session = await get_session(self.request)
        user = get_user_from_session(session, self.request.db_session)
        user.logout(session)
        resp_data = {'success': True}
        return web.json_response(resp_data)



import logging

from aiohttp import web
import aiohttp_jinja2
from aiohttp_session import get_session

from brome.webserver.server.server_decorator import (
    exception_handler
)
from brome.webserver.server.auth import get_user_from_session
from brome.webserver.server.utils import generate_token

logger = logging.getLogger('bromewebserver')


async def set_csrf_token_session(session):
    if session.new:
        session['csrf_token'] = generate_token(20)


@aiohttp_jinja2.template('index.html')
async def index(request):
    logger.debug('index')
    return {}


@exception_handler()
async def api_get_session(request):
    logger.debug('get_session')

    session = await get_session(request)
    await set_csrf_token_session(session)
    data = await request.json()

    success = False
    token = session['csrf_token']
    user = None

    uid = session.get('uid')
    if uid:
        user = get_user_from_session(session, request.db_session)
        session['tz'] = data.get('user_timezone')
        if user.enable:
            context = {
                'user': user,
                'db_session': request.db_session,
                'method': 'read',
                'ws_session': session,
                'queue': request.app.queue
            }

            user = await user.serialize(context)
            success = True
        else:
            user.logout(session)
            user = None

    resp_data = {
        'success': success,
        'user': user,
        'token': token
    }
    return web.json_response(resp_data)



import urllib
import hashlib

import bcrypt
from mongoalchemy.document import Index
from mongoalchemy.fields import (
    EnumField,
    BoolField,
    StringField
)
from validate_email import validate_email

from brome.webserver.server.utils import SafeStringField
from brome.core import exceptions
from brome.model.basemodel import BaseModel

NAME_MIN_LEN = 2  # e.g.: Ed
NAME_MAX_LEN = 60  # e.g.: Hubert Blaine Wolfeschlegelsteinhausenbergerdorff, Sr. # noqa


class User(BaseModel):
    name = SafeStringField(required=True, min_length=NAME_MIN_LEN, max_length=NAME_MAX_LEN)  # noqa
    email = SafeStringField(required=True)
    role = EnumField(StringField(), 'admin', 'user', default="user")
    enable = BoolField(default=True)
    gravatar_url = StringField(default="")
    # settings =
    # social_id =
    # social_provider =

    # PASSWORD
    hashed_password = StringField(required=True)
    salt = StringField(required=True)

    # INDEX
    i_email = Index().ascending('email').unique()

    def __eq__(self, target):
        return target.get_uid() == self.get_uid()

    def __repr__(self):
        try:
            _repr = "User <uid: '{uid}'><name:'{name}'><email:'{email}'><role:'{role}'><enable:'{enable}'>"  # noqa
            return _repr.format(
                    name=self.name,
                    uid=self.get_uid(),
                    email=self.email,
                    enable=self.enable,
                    role=self.role
                )
        except AttributeError:
            return "User uninitialized"

##############################################################################
# FUNC
##############################################################################

    async def check_password(self, target_password):
        target_hashed_password, _ = await self.gen_hashed_password(
            target_password,
            self.salt
        )

        if target_hashed_password == self.hashed_password:
            return True
        else:
            return False

    async def gen_hashed_password(self, raw_password, salt_str=False):
        if not salt_str:
            salt = bcrypt.gensalt()
            salt_str = salt.decode('utf-8')
        else:
            salt = bytes(salt_str.encode('utf-8'))

        mixed_password = raw_password + salt_str
        hashed_password = bcrypt.hashpw(
            mixed_password.encode('utf-8'),
            salt
        ).decode('utf-8')
        return hashed_password, salt_str

    async def sanitize_data(self, context):
        author = context.get('author')
        data = context.get('data')

        if author:
            if author.role == 'admin':
                return data
            else:
                editable_fields = [
                    'name',
                    'email',
                    'old_password',
                    'new_password'
                ]
        else:
            editable_fields = ['name', 'email', 'password']

        return {k: data[k] for k in data if k in editable_fields}

    async def validate_and_save(self, context):
        data = context.get('data')
        db_session = context.get('db_session')
        save = context.get('save', True)

        is_new = await self.is_new()

        # NAME
        name = data.get('name')
        if name:
            if len(name) < NAME_MIN_LEN or len(name) > NAME_MAX_LEN:
                raise exceptions.InvalidNameException(name)

            self.name = name
        else:
            if is_new:
                raise exceptions.InvalidNameException('empty name')

        # ROLE
        role = data.get('role')
        if role:
            self.role = role

        # ENABLE
        enable = data.get('enable')
        if enable is not None:
            self.enable = enable

        # PASSWORD
        password = data.get('password')
        if password:
            await self.set_password(password)
        else:
            if is_new:
                raise exceptions.InvalidPasswordException('empty password')

        # NEW PASSWORD
        new_password = data.get('new_password')
        if new_password:
            old_password = data.get('old_password')
            if old_password:
                is_password_valid = await self.check_password(old_password)
                if is_password_valid:
                    await self.set_password(new_password)
                else:
                    raise exceptions.WrongEmailOrPasswordException()
            else:
                raise exceptions.InvalidRequestException(
                    'Missing old password'
                )

        # EMAIL
        email = data.get('email')
        if email:
            if is_new or self.email != email:
                is_email_valid = validate_email(email)
                if not is_email_valid:
                    raise exceptions.InvalidEmailException(email)

                email_uniqueness_query = db_session.query(User)\
                    .filter(User.email == email)
                if not is_new:
                    email_uniqueness_query = email_uniqueness_query\
                        .filter(User.mongo_id != self.get_uid())

                if email_uniqueness_query.count():
                    raise exceptions.EmailAlreadyExistsException(email)

                self.email = email
                self.email_confirmed = False

                # GRAVATAR
                gravatar_url = "{base_url}{md5_hash}?{params}".format(
                    base_url="https://www.gravatar.com/avatar/",
                    md5_hash=hashlib.md5(
                        email.lower().encode('utf')
                    ).hexdigest(),
                    params=urllib.parse.urlencode(
                        {'d': "identicon", 's': '40'}
                    )
                )
                self.gravatar_url = gravatar_url

        else:
            if is_new:
                raise exceptions.InvalidEmailException('empty email')

        if save:
            db_session.save(self, safe=True)

    async def set_password(self, password):
        if await self.is_password_valid(password):
            hashed_password, salt = await self.gen_hashed_password(password)
            self.hashed_password = hashed_password
            self.salt = salt
        else:
            raise exceptions.InvalidPasswordException(password)

    async def is_password_valid(self, password):
        if len(password) < 6:
            return False
        else:
            return True

    async def method_autorized(self, context):
        method = context.get('method')
        author = context.get('author')

        if method in ['create', 'delete']:
            if author.role == 'admin':
                return True
            else:
                return False
        elif method in ['update', 'read']:
            if author == self:
                return True
            elif author.role == 'admin':
                return True
            else:
                return False

    async def serialize(self, context):
        data = {}
        data['uid'] = self.get_uid()
        data['name'] = self.name
        data['email'] = self.email
        data['gravatar_url'] = self.gravatar_url
        return data

    def logout(self, session):
        del session['uid']



import socket
from time import sleep
import threading
import traceback
import os
from subprocess import Popen
from datetime import datetime
import math

# import virtualbox

from brome.core.utils import (
    DbSessionContext,
    utcnow
)
from brome.core.settings import BROME_CONFIG
from brome.runner.base_runner import BaseRunner
from brome.model.testbatch import Testbatch
from brome.runner import ec2_instance
from brome.runner import virtualbox_instance
from brome.runner import browserstack_instance
from brome.runner import localhost_instance
from brome.runner import saucelabs_instance
from brome.runner.browser_config import BrowserConfig


class GridRunner(BaseRunner):
    """ GridRunner run on Ec2, Virtualbox,
        Browserstack, saucelabs and also localhost
    """
    def __init__(self, *args):
        super(GridRunner, self).__init__(*args)

        self.selenium_pid = None
        self.instances_ip = {}
        self.instances = {}
        self.alive_instances = []
        self.xvfb_pids = []
        self.browser_configs = {}

    def execute(self):
        """Execute the test batch
        """

        try:
            # START SELENIUM GRID
            if BROME_CONFIG['grid_runner']['start_selenium_server']:
                self.start_selenium_server()

            # Get all the browsers id
            self.browsers_id = BROME_CONFIG['runner_args']['remote_runner'].split(',')  # noqa

            # Start all the instances
            instance_threads = []
            for i, browser_id in enumerate(self.browsers_id):
                browser_config = BrowserConfig(
                    runner=self,
                    browser_id=browser_id,
                    browsers_config=BROME_CONFIG['browsers_config']
                )

                self.browser_configs[browser_id] = browser_config

                # EC2
                if browser_config.location == 'ec2':

                    max_number_of_instance = browser_config.get(
                        'max_number_of_instance'
                    )
                    nb_browser_by_instance = browser_config.get(
                        'nb_browser_by_instance'
                    )

                    if len(self.tests) < \
                            max_number_of_instance * \
                            nb_browser_by_instance:

                        nb_instance_to_launch = int(
                            math.ceil(
                                float(len(self.tests))/nb_browser_by_instance
                            )
                        )

                    else:
                        nb_instance_to_launch = max_number_of_instance

                    for j in range(nb_instance_to_launch):
                        _ec2_instance = ec2_instance.EC2Instance(
                            runner=self,
                            browser_config=browser_config,
                            index=j
                        )

                        ec2_instance_thread = InstanceThread(_ec2_instance)
                        ec2_instance_thread.start()

                        instance_threads.append(ec2_instance_thread)

                # VIRTUALBOX
                elif browser_config.location == 'virtualbox':
                    # Instanciate only one vbox
                    """
                    if not hasattr(self, 'vbox'):
                        self.vbox = virtualbox.VirtualBox()
                    """

                    vbox_instance = virtualbox_instance.VirtualboxInstance(
                        runner=self,
                        browser_config=browser_config,
                        index=i,
                    )
                    # vbox=self.vbox

                    vbox_instance_thread = InstanceThread(vbox_instance)
                    vbox_instance_thread.start()

                    instance_threads.append(vbox_instance_thread)

                # SAUCELABS
                elif browser_config.location == 'saucelabs':

                    if not self.instances.get(browser_id):
                        self.instances[browser_id] = []

                    self.instances[browser_id].append(
                        saucelabs_instance.SauceLabsInstance()
                    )

                # BROWSERSTACK
                elif browser_config.location == 'browserstack':

                    if not self.instances.get(browser_id):
                        self.instances[browser_id] = []

                    self.instances[browser_id].append(
                        browserstack_instance.BrowserstackInstance()
                    )

                # LOCALHOST
                elif browser_config.location == 'localhost':

                    max_number_of_instance = browser_config.get(
                        'max_number_of_instance', 1
                    )

                    if len(self.tests) < max_number_of_instance:
                        nb_instance_to_launch = len(self.tests)
                    else:
                        nb_instance_to_launch = max_number_of_instance

                    for i in range(0, nb_instance_to_launch):
                        if not self.instances.get(browser_id):
                            self.instances[browser_id] = []

                        self.instances[browser_id].append(
                            localhost_instance.LocalhostInstance(
                                self,
                                browser_config,
                                test_name=i
                            )
                        )

            # MILESTONE
            if len(instance_threads):
                with DbSessionContext(BROME_CONFIG['database']['mongo_database_name']) as session:  # noqa
                    test_batch = session.query(Testbatch)\
                        .filter(Testbatch.mongo_id == self.test_batch_id).one()
                    test_batch.add_milestone(
                        'NbInstanceToSetup',
                        {'nb': len(instance_threads)}
                    )
                    session.save(test_batch, safe=True)

            for t in instance_threads:
                t.join()

            self.info_log("The test batch is now ready!")

            # MILESTONE
            if len(instance_threads):
                with DbSessionContext(BROME_CONFIG['database']['mongo_database_name']) as session:  # noqa
                    test_batch = session.query(Testbatch)\
                        .filter(Testbatch.mongo_id == self.test_batch_id).one()
                    test_batch.add_milestone('InstanceSetupCompleted')
                    session.save(test_batch, safe=True)

            try:
                self.run()
            except:
                tb = traceback.format_exc()
                self.error_log(
                    "Exception in run of the grid runner: %s"
                    % str(tb)
                )
                raise

            finally:
                try:
                    self.tear_down_instances()

                    # MILESTONE
                    with DbSessionContext(BROME_CONFIG['database']['mongo_database_name']) as session:  # noqa
                        test_batch = session.query(Testbatch)\
                            .filter(Testbatch.mongo_id == self.test_batch_id)\
                            .one()
                        test_batch.add_milestone('InstanceTearDownCompleted')
                        session.save(test_batch, safe=True)

                    # Kill selenium server
                    if BROME_CONFIG['grid_runner']['kill_selenium_server']:  # noqa
                        if self.selenium_pid:
                            self.kill_pid(self.selenium_pid)

                            # MILESTONE
                            with DbSessionContext(BROME_CONFIG['database']['mongo_database_name']) as session:  # noqa
                                test_batch = session.query(Testbatch)\
                                    .filter(Testbatch.mongo_id == self.test_batch_id).one()  # noqa
                                test_batch.add_milestone(
                                    'SeleniumServerKilled'
                                )
                                session.save(test_batch, safe=True)

                    # Kill xvfb process
                    for xvfb_pid in self.xvfb_pids:
                        self.kill_pid(xvfb_pid)

                except:
                    tb = traceback.format_exc()
                    self.error_log("Exception in finally block of the grid runner: %s" % str(tb))  # noqa
        except:
            self.set_ending_timestamp()
            tb = traceback.format_exc()
            self.error_log("Exception in finally block of the grid runner: %s" % str(tb))  # noqa
            raise

        self.info_log("The test batch is finished.")

    def resolve_instance_by_ip(self, ip):
        """Resolve an instance by his ip
        """
        return self.instances_ip[ip]

    def run(self):
        """Run all the test in the test batch
        """

        executed_tests = []
        try:

            active_thread = 0
            start_thread = True
            current_index = 0
            active_thread_by_browser_id = {}

            test_index_by_browser_id = {}
            for browser_id in self.browsers_id:
                test_index_by_browser_id[browser_id] = 0

            self.kill_test_batch_if_necessary()

            while active_thread or start_thread:
                start_thread = False

                self.kill_test_batch_if_necessary()

                for browser_id in self.browsers_id:

                    if active_thread_by_browser_id.get(browser_id) is None:
                        active_thread_by_browser_id[browser_id] = 0
                    else:
                        current_active_thread = 0
                        for thread in threading.enumerate():

                            # Make sure that the thread is TestThread
                            if hasattr(thread, 'test'):
                                if thread.test._browser_config.browser_id == browser_id:  # noqa
                                    current_active_thread += 1

                        active_thread_by_browser_id[browser_id] = current_active_thread  # noqa

                    for j in range(0, len(self.instances[browser_id]) - active_thread_by_browser_id[browser_id]):  # noqa

                        self.kill_test_batch_if_necessary()

                        if test_index_by_browser_id[browser_id] < len(self.tests):  # noqa
                            current_index += 1

                            test = self.tests[test_index_by_browser_id[browser_id]]  # noqa

                            test_index_by_browser_id[browser_id] += 1

                            test_ = test.Test(
                                runner=self,
                                test_batch_id=self.test_batch_id,
                                browser_config=self.browser_configs[browser_id],  # noqa
                                name=test.Test.name,
                                index=test_index_by_browser_id[browser_id]
                            )
                            test_.pdriver.embed_disabled = True

                            thread = TestThread(test_)
                            thread.start()

                            executed_tests.append(test_)

                active_thread = len([tn for tn in threading.enumerate() if type(tn) != threading._MainThread and hasattr(tn, 'test')])  # noqa
                # self.info_log("active_thread=%s" % active_thread)
                if active_thread:
                    try:
                        active_thread_test_number = len([tn for tn in threading.enumerate() if type(tn) != threading._MainThread and hasattr(tn, 'test')])  # noqa
                        self.info_log("Active thread number: %s" % active_thread_test_number)  # noqa
                        self.info_log("Active thread name: %s" % (', '.join([  # noqa
                                "%s-%s" % (
                                    th.test._browser_config.browser_id,
                                    th.test._name
                                ) for th in threading.enumerate() if type(th) != threading._MainThread and hasattr(th, 'test')  # noqa
                            ])
                        ))
                    except Exception as e:
                        self.error_log("print active exception: %s" % str(e))

                # TIMEOUT
                now = utcnow()
                if (self.starting_timestamp - now).total_seconds() >\
                        BROME_CONFIG['grid_runner']['max_running_time']:

                    self.error_log("max_running_time reached... terminating!")
                    raise TestRunnerKilledException()

                self.kill_test_batch_if_necessary()

                sleep(10)

        except TestRunnerKilledException:
            pass

        except Exception:
            tb = traceback.format_exc()
            self.error_log("Run exception: %s" % str(tb))

        self.set_ending_timestamp()

        self.print_test_summary(executed_tests)

    def set_ending_timestamp(self):
        with DbSessionContext(BROME_CONFIG['database']['mongo_database_name']) as session:  # noqa
            test_batch = session.query(Testbatch)\
                .filter(Testbatch.mongo_id == self.test_batch_id).one()
            test_batch.ending_timestamp = utcnow()
            test_batch.terminated = True
            session.save(test_batch, safe=True)

    def kill_test_batch_if_necessary(self):
        """Kill the test batch

            If the test_batch.killed is set to true
            then the test batch will be kill
        """
        with DbSessionContext(BROME_CONFIG['database']['mongo_database_name']) as session:  # noqa
            test_batch = session.query(Testbatch)\
                .filter(Testbatch.mongo_id == self.test_batch_id).one()
            if test_batch.killed:
                test_batch.add_milestone('TerminatingTestBatch')
                session.save(test_batch, safe=True)

                self.info_log("Killing itself")
                for t in [t for t in threading.enumerate() if type(t) != threading._MainThread]:  # noqa
                    if not t.test.ending_timestamp:
                        self.info_log("Killing: %s" % t.test._name)
                        t.test.end()

                raise TestRunnerKilledException("Killed")

    def tear_down_instances(self):
        """Tear down all instances
        """

        self.info_log('Tearing down all instances...')

        for instance in self.alive_instances:
            instance.tear_down()

        self.info_log('[Done]Tearing down all instances')

    def start_selenium_server(self):
        """Start the selenium server
        """

        ip = BROME_CONFIG['grid_runner']['selenium_server_ip']
        port = BROME_CONFIG['grid_runner']['selenium_server_port']

        def is_selenium_server_is_running():
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            result = s.connect_ex((ip, port))
            s.close()
            return not result

        if not is_selenium_server_is_running():
            self.info_log('Starting selenium server...')

            command = BROME_CONFIG['grid_runner']['selenium_server_command']\
                .format(
                    **BROME_CONFIG['grid_runner']
                )

            self.info_log('Selenium hub command: %s' % command)

            process = Popen(
                command.split(' '),
                stdout=open(os.path.join(self.runner_dir, "hub.log"), 'a'),
                stderr=open(os.path.join(self.runner_dir, "hub.log"), 'a'),
            )

            self.selenium_pid = process.pid

            self.info_log('Selenium server pid: %s' % self.selenium_pid)
        else:
            self.info_log('Selenium is already running.')

            # Milestone
            with DbSessionContext(BROME_CONFIG['database']['mongo_database_name']) as session:  # noqa
                test_batch = session.query(Testbatch)\
                    .filter(Testbatch.mongo_id == self.test_batch_id).one()  # noqa
                test_batch.add_milestone(
                    'SeleniumServerAlreadyRunning'
                )
                session.save(test_batch, safe=True)

            return True

        for i in range(30):
            self.info_log('Waiting for the selenium server to start...')
            result = is_selenium_server_is_running()

            if result:
                self.info_log('[Done]Selenium server is running.')

                # Milestone
                with DbSessionContext(BROME_CONFIG['database']['mongo_database_name']) as session:  # noqa
                    test_batch = session.query(Testbatch)\
                        .filter(Testbatch.mongo_id == self.test_batch_id).one()  # noqa
                    test_batch.add_milestone(
                        'SeleniumServerStarted'
                    )
                    session.save(test_batch, safe=True)

                return True
            sleep(2)

        raise Exception("Selenium server did not start!")


class InstanceThread(threading.Thread):
    """Theard that start the instance
    """

    def __init__(self, instance):
        threading.Thread.__init__(self)
        self.instance = instance
        self.runner = self.instance.runner

    def run(self):
        success = False
        try:
            success = self.instance.startup()
        except Exception as e:
            self.runner.critical_log(
                "Exception in InstanceThread instance startup: %s" % e
            )

        if not self.runner.instances.get(self.instance.browser_config.browser_id):  # noqa
            self.runner.instances[self.instance.browser_config.browser_id] = []

        if success:
            self.runner.instances[self.instance.browser_config.browser_id].append(self.instance)  # noqa
            self.runner.instances_ip[self.instance.get_ip()] = self.instance

        self.runner.alive_instances.append(self.instance)


class TestThread(threading.Thread):
    """Thread that run the test
    """

    def __init__(self, test):
        threading.Thread.__init__(self)
        self.test = test

    def run(self):
        self.test.execute()


class TestRunnerKilledException(Exception):
    """Exception that is raised when the test batch is killed
    """
    pass



# from abc import ABCMeta, abstractmethod

from mongoalchemy.document import Document
from mongoalchemy.fields import (
    ModifiedField,
    CreatedField
)


# TODO make it a meta class
class BaseModel(Document):

    # TIMESTAMP
    created_ts = CreatedField(tz_aware=True)
    modified_ts = ModifiedField(tz_aware=True)

    # @abstractmethod
    async def sanitize_data(self, context):
        raise NotImplementedError()

    # @abstractmethod
    async def serialize(self, context):
        raise NotImplementedError()

    # @abstractmethod
    async def method_autorized(self, context):
        raise NotImplementedError()

    # @abstractmethod
    async def validate_and_save(self, context):
        raise NotImplementedError()

    async def is_new(self):
        return not hasattr(self, 'mongo_id')

    def get_uid(self):
        return str(self.mongo_id)

    async def before_create(self, context):
        pass

    async def after_create(self, context):
        pass

    async def before_update(self, context):
        pass

    async def after_update(self, context):
        pass

    async def before_delete(self, context):
        pass

    async def after_delete(self, context):
        pass



from datetime import datetime
import os
import shutil

from mongoalchemy.fields import (
    DateTimeField,
    DictField,
    AnythingField,
    StringField,
    IntField,
    BoolField
)

from brome.core.utils import convert_tz_datetime
from brome.model.basemodel import BaseModel
from brome.model.testinstance import Testinstance
from brome.model.testcrash import Testcrash
from brome.model.testresult import Testresult
from brome.model.testscreenshot import Testscreenshot
from brome.core import exceptions
from brome.core.settings import BROME_CONFIG


class Testbatch(BaseModel):

    friendly_name = StringField()
    pid = IntField()
    killed = BoolField(default=False)
    total_tests = IntField()
    starting_timestamp = DateTimeField(use_tz=True)
    ending_timestamp = DateTimeField(required=False, use_tz=True)
    terminated = BoolField(default=False)
    feature_session_video_capture = BoolField(default=False)
    feature_network_capture = BoolField(default=False)
    feature_bot_diaries = BoolField(default=False)
    feature_screenshots = BoolField(default=False)
    feature_instance_vnc = BoolField(default=False)
    feature_style_quality = BoolField(default=False)
    runner_metadata = DictField(AnythingField(), default=dict())
    root_path = StringField(default='')
    log_file_path = StringField(default='')

    def __repr__(self):
        try:
            return "Testbatch <uid: {self.mongo_id}><friendly_name: {self.friendly_name}>".format(  # noqa
                self=self
            )
        except AttributeError:
            return "Testbatch uninitialized"

    def add_milestone(self, milestone_id, values=None, index=None):
        if 'milestones' not in self.runner_metadata.keys():
            self.runner_metadata['milestones'] = {}

        effective_index = index
        effective_values = values
        if not index:
            effective_index = len(self.runner_metadata['milestones'])

        if not values:
            effective_values = {}

        self.runner_metadata['milestones'][milestone_id] = {
            'msg_id': milestone_id,
            'index': effective_index,
            'values': effective_values
        }

    def update_milestone(self, milestone_id, values):
        self.runner_metadata['milestones'][milestone_id]['values'] = values

    async def get_browser_ids(self, context):
        db_session = context.get('db_session')

        browser_ids = []

        unique_browser_ids = db_session.query(Testinstance)\
            .filter(Testinstance.test_batch_id == self.get_uid())\
            .distinct('browser_id')

        for browser_id in unique_browser_ids:
            capabilities = db_session.query(Testinstance)\
                .filter(Testinstance.test_batch_id == self.get_uid())\
                .filter(Testinstance.browser_id == browser_id)\
                .first().browser_capabilities

            browser_id_dict = {
                'id': browser_id,
                'capabilities': capabilities
            }
            browser_ids.append(browser_id_dict)

        return browser_ids

    async def get_milestones(self):
        data = []
        if 'milestones' in self.runner_metadata.keys():
            sorted_milestones = sorted(
                self.runner_metadata['milestones'].values(),
                key=lambda x: x['index']
            )
            for milestone in sorted_milestones:
                data.append({
                    'values': milestone['values'],
                    'msgId': milestone['msg_id']
                })

        return data

    async def get_test_results(self, context):
        db_session = context.get('db_session')

        nb_test_result = db_session.query(Testresult)\
            .filter(Testresult.test_batch_id == self.get_uid())\
            .count()

        failed_tests_query = db_session.query(Testresult)\
            .filter(Testresult.test_batch_id == self.get_uid())\
            .filter(Testresult.result == False)  # noqa

        succeeded_test_query = db_session.query(Testresult)\
            .filter(Testresult.test_batch_id == self.get_uid())\
            .filter(Testresult.result == True)  # noqa

        data = {}
        data['nb_failed_test'] = failed_tests_query.count()
        data['nb_succeeded_test'] = succeeded_test_query.count()
        data['nb_test_result'] = nb_test_result
        data['failed_tests'] = []

        failed_tests_title = []
        for failed_test in failed_tests_query.all():
            if failed_test.title not in failed_tests_title:
                failed_tests_title.append(failed_test.title)
                data['failed_tests'].append(
                    await failed_test.serialize(context)
                )

        return data

    async def get_total_executings_tests(self, context):
        db_session = context.get('db_session')

        return db_session.query(Testinstance)\
            .filter(Testinstance.terminated == False)\
            .filter(Testinstance.test_batch_id == self.mongo_id)\
            .count()  # noqa

    async def get_total_executed_tests(self, context):
        db_session = context.get('db_session')

        return db_session.query(Testinstance)\
            .filter(Testinstance.terminated == True)\
            .filter(Testinstance.test_batch_id == self.mongo_id)\
            .count()  # noqa

    async def get_test_crashes(self, context):
        db_session = context.get('db_session')

        data = []

        test_crashes = db_session.query(Testcrash)\
            .filter(Testcrash.test_batch_id == self.get_uid())\
            .ascending('title')\
            .all()

        # TODO auto serialize
        for test_crash in test_crashes:
            data.append(await test_crash.serialize(context))

        return data

    async def get_nb_screenshot(self, context):
        db_session = context.get('db_session')

        query = db_session.query(Testscreenshot)\
            .filter(Testscreenshot.test_batch_id == self.mongo_id)

        return query.count()

    async def sanitize_data(self, context):
        author = context.get('author')
        data = context.get('data')

        if author:
            if author.role == 'admin':
                return data
            else:
                editable_fields = [
                    'killed'
                ]
        else:
            editable_fields = []

        return {k: data[k] for k in data if k in editable_fields}

    async def serialize(self, context):
        ws_session = context.get('ws_session')

        data = {}
        data['uid'] = self.get_uid()
        data['friendly_name'] = self.friendly_name
        data['killed'] = self.killed
        data['total_tests'] = self.total_tests
        data['browser_ids'] = await self.get_browser_ids(context)
        data['total_executed_tests'] = await self.get_total_executed_tests(context)  # noqa
        data['total_executing_tests'] = await self.get_total_executings_tests(context)  # noqa
        data['starting_timestamp'] = convert_tz_datetime(
            self.starting_timestamp,
            ws_session['tz']
        ).isoformat()
        data['nb_screenshot'] = await self.get_nb_screenshot(context)
        data['root_path'] = self.root_path
        data['log_file_path'] = self.log_file_path
        data['features'] = {
            'session_video_capture': self.feature_session_video_capture,
            'network_capture': self.feature_network_capture,
            'bot_diaries': self.feature_bot_diaries,
            'screenshots': self.feature_screenshots,
            'instance_vnc': self.feature_instance_vnc,
            'style_quality': self.feature_style_quality
        }
        data['milestones'] = await self.get_milestones()

        data['test_crashes'] = await self.get_test_crashes(context)
        data['test_results'] = await self.get_test_results(context)

        data['terminated'] = self.terminated
        # Terminated test batch
        if self.terminated:
            data['ending_timestamp'] = convert_tz_datetime(
                self.ending_timestamp,
                ws_session['tz']
            ).isoformat()

        # Running test batch
        else:
            data['ending_timestamp'] = False

        return data

    async def method_autorized(self, context):
        author = context.get('author')
        method = context.get('method')

        if method in ['create']:
            return False
        else:
            if author:
                return True
            else:
                return False

    async def validate_and_save(self, context):
        data = context.get('data')
        db_session = context.get('db_session')

        is_new = await self.is_new()

        # Killed
        killed = data.get('killed')
        if killed is not None:
            self.killed = killed

        # PID
        pid = data.get('pid')
        if pid is not None:
            self.pid = pid

        # TOTAL TESTS
        total_tests = data.get('total_tests')
        if total_tests is not None:
            self.total_tests = total_tests
        else:
            if is_new:
                raise exceptions.MissingModelValueException('total_tests')

        # FRIENDLY NAME
        friendly_name = data.get('friendly_name')
        if friendly_name:
            self.friendly_name = friendly_name
        else:
            if is_new:
                raise exceptions.MissingModelValueException('friendly_name')

        # STARTING TIMESTAMP
        starting_timestamp = data.get('starting_timestamp')
        if starting_timestamp:
            self.starting_timestamp = starting_timestamp
        else:
            if is_new:
                self.starting_timestamp = datetime.now()

        # ENDING TIMESTAMP
        ending_timestamp = data.get('ending_timestamp')
        if ending_timestamp:
            self.ending_timestamp = ending_timestamp

        # FEATURES
        feature_session_video_capture = data.get(
            'feature_session_video_capture'
        )
        if feature_session_video_capture:
            self.feature_session_video_capture = feature_session_video_capture

        feature_network_capture = data.get('feature_network_capture')
        if feature_network_capture:
            self.feature_network_capture = feature_network_capture

        feature_bot_diaries = data.get('feature_bot_diaries')
        if feature_bot_diaries:
            self.feature_bot_diaries = feature_bot_diaries

        feature_screenshots = data.get('feature_screenshots')
        if feature_screenshots:
            self.feature_screenshots = feature_screenshots

        # RUNNER METADATA
        runner_metadata = data.get('runner_metadata')
        if runner_metadata:
            self.runner_metadata = runner_metadata

        # LOG FILE PATH
        log_file_path = data.get('log_file_path')
        if log_file_path:
            self.log_file_path = log_file_path

        # ROOT PATH
        root_path = data.get('root_path')
        if root_path:
            self.root_path = root_path

        db_session.save(self, safe=True)

    async def after_delete(self, context):
        if BROME_CONFIG['webserver']['env'] != 'test':
            root_path = os.path.join(
                self.root_path,
                os.sep.join(self.log_file_path.split(os.sep)[:-1])
            )
            shutil.rmtree(root_path)



def generate_brome_config():
    """Generate a brome config with default value

    Returns:
        config (dict)
    """

    config = {}
    for key in iter(default_config):
        for inner_key, value in iter(default_config[key].items()):
            if key not in config:
                config[key] = {}

            config[key][inner_key] = value['default']

    return config


def test_config_to_dict(test_config_string):
    """Parse the test config to a dictionary

    Args:
        test_config_string (str) this string come from the --test-config
        flag of the bro executable run command
    """

    test_config = {}
    if test_config_string:
        for config in test_config_string.split(','):
            key, value = config.split('=')
            test_config[key] = value

    return test_config


def parse_brome_config_from_browser_config(browser_config):
    """Parse the browser config and look for brome specific config

    Args:
        browser_config (dict)
    """

    config = {}

    brome_keys = [key for key in browser_config if key.find(':') != -1]

    for brome_key in brome_keys:
        section, option = brome_key.split(':')
        value = browser_config[brome_key]

        if section not in config:
            config[section] = {}

        config[section][option] = value

    return config

default_config = {}

# ##SECTIONS
default_config["mitmproxy"] = {}
default_config["brome"] = {}
default_config["project"] = {}
default_config["saucelabs"] = {}
default_config["browserstack"] = {}
default_config["proxy_driver"] = {}
default_config["proxy_element"] = {}
default_config["browser"] = {}
default_config["highlight"] = {}
default_config["runner"] = {}
default_config["database"] = {}
default_config["logger_runner"] = {}
default_config["bot_diary"] = {}
default_config["logger_test"] = {}
default_config["ec2"] = {}
default_config["grid_runner"] = {}
default_config["webserver"] = {}
default_config["runner_args"] = {}

# RUNNER ARGS
default_config["runner_args"]["remote_runner"] = {
    'default': '',
    'type': 'input',
    'visible': False,
    'title': ''
}

default_config["runner_args"]["localhost_runner"] = {
    'default': '',
    'type': 'input',
    'visible': False,
    'title': ''
}

# MITMPROXY
default_config["mitmproxy"]["path"] = {
    'default': '',
    'type': 'input',
    'visible': False,
    'title': 'The path to the mitmproxy python bin'
}

default_config["mitmproxy"]["filter"] = {
    'default': '',
    'type': 'input',
    'visible': False,
    'title': 'The mitmproxy filter'
}

# BROME
default_config["brome"]["script_folder_name"] = {
    'default': 'tests',
    'type': 'input',
    'visible': False,
    'title': 'The name of the folder holding the scripts'
}

default_config["brome"]["script_test_prefix"] = {
    'default': 'test_',
    'type': 'input',
    'visible': False,
    'title': 'The prefix of the script'
}

default_config["brome"]["brome_executable_name"] = {
    'default': 'bro',
    'type': 'input',
    'visible': False,
    'title': 'The brome executable name'
}

# PROXY ELEMENT
default_config["proxy_element"]["use_touch_instead_of_click"] = {
    'default': False,
    'type': 'checkbox',
    'visible': True,
    'title': 'Use touch instead of click'
}

default_config["proxy_element"]["wait_until_clickable"] = {
    'default': True,
    'type': 'checkbox',
    'visible': True,
    'title': 'Wait until clickable'
}

# BROWSERSTACK
default_config["browserstack"]["username"] = {
    'default': '',
    'type': 'input',
    'visible': True,
    'title': 'Browserstack username'
}

default_config["browserstack"]["key"] = {
    'default': '',
    'type': 'input',
    'visible': True,
    'title': 'Browserstack key'
}

# SAUCELABS
default_config["saucelabs"]["username"] = {
    'default': '',
    'type': 'input',
    'visible': True,
    'title': 'Saucelabs username'
}

default_config["saucelabs"]["key"] = {
    'default': '',
    'type': 'input',
    'visible': True,
    'title': 'Saucelabs key'
}

# GRID RUNNER
default_config["grid_runner"]["max_running_time"] = {
    'default': 7200,
    'type': 'number',
    'visible': True,
    'title': 'Max running time'
}

default_config["grid_runner"]["start_selenium_server"] = {
    'default': True,
    'type': 'checkbox',
    'visible': True,
    'title': 'Start selenium server automatically'
}

default_config["grid_runner"]["selenium_server_ip"] = {
    'default': 'localhost',
    'type': 'input',
    'visible': True,
    'title': 'Selenium server ip address'
}

default_config["grid_runner"]["selenium_server_port"] = {
    'default': 4444,
    'type': 'number',
    'visible': True,
    'title': 'Selenium port'
}
default_config["grid_runner"]["selenium_server_command"] = {
    'default': "",
    'type': 'input',
    'visible': True,
    'title': 'Selenium server command'
}

default_config["grid_runner"]["selenium_server_jar_path"] = {
    'default': "",
    'type': 'input',
    'visible': True,
    'title': 'Selenium server jar path'
}

default_config["grid_runner"]["selenium_hub_config"] = {
    'default': "",
    'type': 'input',
    'visible': True,
    'title': 'Selenium server hub config path'
}

default_config["grid_runner"]["kill_selenium_server"] = {
    'default': True,
    'type': 'checkbox',
    'visible': True,
    'title': 'Kill selenium server when the test batch finished'
}

# EC2
default_config["ec2"]['wait_after_instance_launched'] = {
    'default': 30,
    'type': 'number',
    'visible': True,
    'title': 'Wait X seconds after the instances are launched'
}

default_config["ec2"]['wait_until_system_and_instance_check_performed'] = {
    'default': True,
    'type': 'checkbox',
    'visible': True,
    'title': 'Wait until system and instance checks are performed'
}

# PROJECT
default_config["project"]["script_folder_name"] = {
    'default': "tests",
    'type': 'input',
    'visible': True,
    'title': 'The script folder name'
}

default_config["project"]["test_batch_result_path"] = {
    'default': "",
    'type': 'input',
    'visible': True,
    'title': 'Test batch result path'
}

default_config["project"]["url"] = {
    'default': "",
    'type': 'input',
    'visible': True,
    'title': 'The url of the server on which the test run (must include the protocol) e.g.:https://the-internet.herokuapp.com/'  # noqa
}

# LOGGER RUNNER
default_config["logger_runner"]["level"] = {
    'default': "INFO",
    'type': 'dropdown',
    'options': ['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
    'visible': True,
    'title': 'Logger level'
}

default_config["logger_runner"]["streamlogger"] = {
    'default': True,
    'type': 'checkbox',
    'visible': True,
    'title': 'Use stream logger'
}

default_config["logger_runner"]["filelogger"] = {
    'default': True,
    'type': 'checkbox',
    'visible': True,
    'title': 'Use file logger'
}

default_config["logger_runner"]["format"] = {
    'default': "[%(batchid)s]%(message)s",
    'type': 'input',
    'visible': True,
    'title': 'Logger format'
}

# LOGGER TEST
default_config["logger_test"]["level"] = {
    'default': "INFO",
    'type': 'dropdown',
    'options': ['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
    'visible': True,
    'title': 'Logger level'
}

default_config["logger_test"]["streamlogger"] = {
    'default': True,
    'type': 'checkbox',
    'visible': True,
    'title': 'Use stream logger'
}

default_config["logger_test"]["filelogger"] = {
    'default': True,
    'type': 'checkbox',
    'visible': True,
    'title': 'Use file logger'
}

default_config["logger_test"]["format"] = {
    'default': "[%(batchid)s](%(testname)s):%(message)s",
    'type': 'input',
    'visible': True,
    'title': 'Logger format'
}

# PROXY DRIVER
default_config["proxy_driver"]["use_javascript_dnd"] = {
    'default': False,
    'type': 'checkbox',
    'visible': True,
    'title': 'Use javascript to perform drag and drop'
}

default_config["proxy_driver"]["wait_until_visible_before_find"] = {
    'default': False,
    'type': 'checkbox',
    'visible': True,
    'title': 'Wait until visible before find'
}

default_config["proxy_driver"]["intercept_javascript_error"] = {
    'default': False,
    'type': 'checkbox',
    'visible': True,
    'title': 'Intercept javascript error'
}

default_config["proxy_driver"]["validate_xpath_selector"] = {
    'default': True,
    'type': 'checkbox',
    'visible': True,
    'title': 'Validate xpath selector'
}

default_config["proxy_driver"]["validate_css_selector"] = {
    'default': True,
    'type': 'checkbox',
    'visible': True,
    'title': 'Validate css selector'
}

default_config["proxy_driver"]["default_timeout"] = {
    'default': 5,
    'type': 'number',
    'visible': True,
    'title': 'Default timeout'
}

default_config["proxy_driver"]["raise_exception"] = {
    'default': True,
    'type': 'checkbox',
    'visible': True,
    'title': 'Raise exception'
}

default_config["proxy_driver"]["wait_until_present_before_assert_present"] = {
    'default': False,
    'type': 'checkbox',
    'visible': True,
    'title': 'Wait until not present before assert present'
}

default_config["proxy_driver"]["wait_until_not_present_before_assert_not_present"] = {  # noqa
    'default': False,
    'type': 'checkbox',
    'visible': True,
    'title': 'Wait until not present  before assert not present'
}

default_config["proxy_driver"]["wait_until_not_visible_before_assert_not_visible"] = {  # noqa
    'default': False,
    'type': 'checkbox',
    'visible': True,
    'title': 'Wait until not visible before assert not visible'
}

default_config["proxy_driver"]["wait_until_visible_before_assert_visible"] = {
    'default': False,
    'type': 'checkbox',
    'visible': True,
    'title': 'Wait until visible before assert visible'
}

default_config["proxy_driver"]["wait_until_present_before_find"] = {
    'default': False,
    'type': 'checkbox',
    'visible': True,
    'title': 'Wait until visible before find'
}

default_config["proxy_driver"]["take_screenshot_on_assertion_success"] = {
    'default': False,
    'type': 'checkbox',
    'visible': True,
    'title': 'Take screenshot on assertion success'
}

default_config["proxy_driver"]["take_screenshot_on_assertion_failure"] = {
    'default': False,
    'type': 'checkbox',
    'visible': True,
    'title': 'Take screenshot on assertion failure'
}

# BROWSER
default_config["browser"]["window_x_position"] = {
    'default': 0,
    'type': 'number',
    'visible': True,
    'title': 'Window x position'
}

default_config["browser"]["window_y_position"] = {
    'default': 0,
    'type': 'number',
    'visible': True,
    'title': 'Window y position'
}

default_config["browser"]["window_height"] = {
    'default': 725,
    'type': 'number',
    'visible': True,
    'title': 'Window height'
}

default_config["browser"]["window_width"] = {
    'default': 1650,
    'type': 'number',
    'visible': True,
    'title': 'Window width'
}

default_config["browser"]["maximize_window"] = {
    'default': False,
    'type': 'checkbox',
    'visible': True,
    'title': 'Maximize window'
}

# HIGHLIGHT
default_config["highlight"]["highlight_on_assertion_success"] = {
    'default': False,
    'type': 'checkbox',
    'visible': True,
    'title': 'Highlight on assertion success'
}

default_config["highlight"]["highlight_on_assertion_failure"] = {
    'default': False,
    'type': 'checkbox',
    'visible': True,
    'title': 'Highlight on assertion failure'
}

default_config["highlight"]["highlight_when_element_is_clicked"] = {
    'default': False,
    'type': 'checkbox',
    'visible': True,
    'title': 'Highlight when element is clicked'
}

default_config["highlight"]["highlight_when_element_receive_keys"] = {
    'default': False,
    'type': 'checkbox',
    'visible': True,
    'title': 'Highlight when element received keys'
}

default_config["highlight"]["element_is_visible"] = {
    'default': False,
    'type': 'checkbox',
    'visible': True,
    'title': 'Highlight when element is visible'
}

default_config["highlight"]["style_when_element_is_clicked"] = {
    'default': "background: yellow; border: 2px solid red;",
    'type': 'input',
    'visible': True,
    'title': 'Element is clicked'
}

default_config["highlight"]["style_when_element_receive_keys"] = {
    'default': "background: yellow; border: 2px solid red;",
    'type': 'input',
    'visible': True,
    'title': 'Element receive keys'
}

default_config["highlight"]["style_on_assertion_failure"] = {
    'default': "background: red; border: 2px solid black;",
    'type': 'input',
    'visible': True,
    'title': 'On assertion failure'
}

default_config["highlight"]["style_on_assertion_success"] = {
    'default': "background: green; border: 2px solid black;",
    'type': 'input',
    'visible': True,
    'title': 'On assertion success style'
}

default_config["highlight"]["style_when_element_is_visible"] = {
    'default': "background: purple; border: 2px solid black;",
    'type': 'input',
    'visible': True,
    'title': 'Element is visible style'
}

default_config["highlight"]["use_highlight"] = {
    'default': False,
    'type': 'checkbox',
    'visible': True,
    'title': 'Use highlight'
}

# RUNNER
default_config["runner"]["embed_on_assertion_success"] = {
    'default': False,
    'type': 'checkbox',
    'visible': True,
    'title': 'Embed on assertion success'
}

default_config["runner"]["embed_on_assertion_failure"] = {
    'default': False,
    'type': 'checkbox',
    'visible': True,
    'title': 'Embed on assertion failure'
}

default_config["runner"]["embed_on_test_crash"] = {
    'default': False,
    'type': 'checkbox',
    'visible': True,
    'title': 'Embed on test crash'
}

default_config["runner"]["play_sound_on_test_crash"] = {
    'default': False,
    'type': 'checkbox',
    'visible': True,
    'title': 'Play sound on test crash'
}

default_config["runner"]["play_sound_on_assertion_success"] = {
    'default': False,
    'type': 'checkbox',
    'visible': True,
    'title': 'Play sound on assertion success'
}

default_config["runner"]["play_sound_on_assertion_failure"] = {
    'default': False,
    'type': 'checkbox',
    'visible': True,
    'title': 'Play sound on assertion failure'
}

default_config["runner"]["play_sound_on_test_finished"] = {
    'default': False,
    'type': 'checkbox',
    'visible': True,
    'title': 'Play sound on test batch finished'
}

default_config["runner"]["play_sound_on_ipython_embed"] = {
    'default': False,
    'type': 'checkbox',
    'visible': True,
    'title': 'Play sound on ipython embed'
}

default_config["runner"]["play_sound_on_pdb"] = {
    'default': False,
    'type': 'checkbox',
    'visible': True,
    'title': 'Play sound on pdb'
}

default_config["runner"]["sound_on_test_crash"] = {
    'default': 'Crash',
    'type': 'input',
    'visible': True,
    'title': 'Sound on test crash'
}

default_config["runner"]["sound_on_assertion_success"] = {
    'default': "{testid} succeeded",
    'type': 'input',
    'visible': True,
    'title': 'sound on assertion success'
}

default_config["runner"]["sound_on_assertion_failure"] = {
    'default': '{testid} failed',
    'type': 'input',
    'visible': True,
    'title': 'Sound on assertion failure'
}

default_config["runner"]["sound_on_test_finished"] = {
    'default': 'Test finished',
    'type': 'input',
    'visible': True,
    'title': 'Sound on test batch finished'
}

default_config["runner"]["sound_on_ipython_embed"] = {
    'default': 'Attention required',
    'type': 'input',
    'visible': True,
    'title': 'Sound on ipython embed'
}

default_config["runner"]["sound_on_pdb"] = {
    'default': 'Attention required',
    'type': 'input',
    'visible': True,
    'title': 'Sound on pdb'
}

default_config["runner"]["cache_screenshot"] = {
    'default': True,
    'type': 'checkbox',
    'visible': True,
    'title': 'Use the cache screenshot'
}

# DATABASE
default_config["database"]["mongo_database_name"] = {
    'default': '',
    'type': 'input',
    'visible': True,
    'title': 'Mongo database name'
}

# BOT DIARY
default_config["bot_diary"]["enable_auto_bot_diary"] = {
    'default': False,
    'type': 'checkbox',
    'visible': True,
    'title': 'Enable auto bot diary'
}

default_config["bot_diary"]["enable_auto_bot_diary_component_screenshot"] = {
    'default': False,
    'type': 'checkbox',
    'visible': True,
    'title': 'Enable auto bot diary component screenshot'
}

default_config["bot_diary"]["enable_bot_diary"] = {
    'default': False,
    'type': 'checkbox',
    'visible': True,
    'title': 'Enable bot diary'
}

default_config["bot_diary"]["streamlogger"] = {
    'default': False,
    'type': 'checkbox',
    'visible': True,
    'title': 'Use stream logger'
}

default_config["bot_diary"]["filelogger"] = {
    'default': True,
    'type': 'checkbox',
    'visible': True,
    'title': 'Use file logger'
}

# WEBSERVER
default_config["webserver"]["LOG_LEVEL"] = {
    'default': 'INFO',
    'type': 'dropdown',
    'options': ['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
    'visible': True,
    'title': 'Logger level'
}

default_config["webserver"]["STREAMLOGGER"] = {
    'default': True,
    'type': 'checkbox',
    'visible': True,
    'title': 'Use stream logger'
}

default_config["webserver"]["FILELOGGER"] = {
    'default': True,
    'type': 'checkbox',
    'visible': True,
    'title': 'Use file logger'
}

default_config["webserver"]["ENV"] = {
    'default': 'production',
    'type': 'checkbox',
    'visible': True,
    'title': 'Environnemnt for the server: "production", "test", "development"',  # noqa
}

default_config["webserver"]["CLOSED_REGISTRATION"] = {
    'default': False,
    'type': 'checkbox',
    'visible': True,
    'title': 'closed registration',
}

default_config["webserver"]["REGISTRATION_TOKEN"] = {
    'default': '',
    'type': 'input',
    'visible': True,
    'title': 'registration token',
}

default_config["webserver"]["PORT"] = {
    'default': 5000,
    'type': 'number',
    'visible': True,
    'title': 'webserver port',
}

default_config["webserver"]["SHOW_QUALITY_SCREENSHOT"] = {
    'default': False,
    'type': 'checkbox',
    'visible': True,
    'title': 'Show the quality screenshot in the detail of a test batch',
}

default_config["webserver"]["SHOW_NETWORK_CAPTURE"] = {
    'default': False,
    'type': 'checkbox',
    'visible': True,
    'title': 'Show the network capture in the detail of a test batch',
}

default_config["webserver"]["SHOW_BOT_DIARY"] = {
    'default': False,
    'type': 'checkbox',
    'visible': True,
    'title': 'Show the bot diary in the detail of a test batch',
}

default_config["webserver"]["SHOW_VIDEO_CAPTURE"] = {
    'default': False,
    'type': 'checkbox',
    'visible': True,
    'title': 'Show the video capture in the detail of a test batch',
}

default_config["webserver"]["SHOW_TEST_INSTANCES"] = {
    'default': False,
    'type': 'checkbox',
    'visible': True,
    'title': 'Show the test instance in the detail of a test batch',
}

default_config["webserver"]["HOST"] = {
    'default': 'localhost',
    'type': 'input',
    'visible': True,
    'title': 'webserver ip',
}

default_config["webserver"]["report"] = {
    'default': False,
    'type': 'input',
    'visible': False,
    'title': 'Config for the report system',
}



import functools
import logging
import traceback

from aiohttp import web
from aiohttp.abc import AbstractView
from aiohttp_session import get_session

from brome.webserver.server.auth import permits
from brome.core import exceptions

logger = logging.getLogger('bromewebserver')


def require(permission):
    def wrapper(func):
        @functools.wraps(func)
        async def wrapped(*args):
            logger.debug('require: {permission}'.format(permission=permission))

            # Supports class based views see web.View
            if isinstance(args[0], AbstractView):
                request = args[0].request
                params = args[0]  # self
            else:
                request = args[-1]
                params = request  # request

            session = await get_session(request)
            has_perm = permits(request, session, permission)
            if not has_perm:
                raise exceptions.NotAuthorizedException(permission)

            return (await func(params))
        return wrapped
    return wrapper


def exception_handler():
    def wrapper(func):
        @functools.wraps(func)
        async def wrapped(*args):
            try:
                # Supports class based views see web.View
                if isinstance(args[0], AbstractView):
                    return (await func(*args))
                else:
                    return (await func(args[-1]))
            except exceptions.CSRFMismatch as e:
                data = {'success': False, 'error': 'CSRFMismatch'}

                return web.json_response(data)
            except Exception as e:
                tb = traceback.format_exc()
                logger.error(
                    'Request HandledException<{exception}>'
                    .format(exception=str(tb))
                )
                if isinstance(e, exceptions.ServerBaseException):
                    data = {'success': False, 'error': e.get_name()}
                else:
                    data = {'success': False, 'error': 'ServerSideError'}

                return web.json_response(data)

        return wrapped
    return wrapper


def csrf_protected():
    def wrapper(func):
        @functools.wraps(func)
        async def wrapped(*args):
            logger.debug('csrf_protected')

            # Supports class based views see web.View
            if isinstance(args[0], AbstractView):
                request = args[0].request
                params = args[0]  # self
            else:
                request = args[-1]
                params = request  # request

            session = await get_session(request)
            try:
                data = await request.json()
            except:
                raise exceptions.InvalidRequestException('No json send')

            csrf_token_session = session.get('csrf_token')
            csrf_token_request = data.get('token')
            if csrf_token_request != csrf_token_session:
                raise exceptions.CSRFMismatch()

            return (await func(params))
        return wrapped
    return wrapper



from brome.core.proxy_element import ProxyElement

class ProxyElementList(list):
    def __init__(self, elements, selector, pdriver):
        self._elements = elements
        self._selector = selector
        self.pdriver = pdriver

    def __getslice__(self, i, j):
        return self.__getitem__(slice(i, j))

    def __reversed__(self):
        return self.__getitem__(slice(None, None, -1))

    def __len__(self):
        return len(self._elements)

    def __iter__(self):
        for element in self._elements:
            yield ProxyElement(element, self._selector, self.pdriver)

    def __getitem__(self, index):
        if isinstance(index, slice):
            return ProxyElementList(self._elements[index.start: index.stop: index.step], self._selector, self.pdriver)
        else:
            return ProxyElement(self._elements[index], self._selector, self.pdriver)

    def __repr__(self):
        if len(self) <= 5:
            return 'WebElement list [\n %s \n]'%',\n '.join([repr(el) for el in self])
        else:
            return 'WebElement list containing %d webelement'%len(self)




class InitDriverException(Exception):
    """Raise when the driver failed initializing
    Args:
        msg (str): Human readable string describing the exception.
    """

    def __init__(self, msg):
        self.msg = msg


class InvalidBrowserName(Exception):
    """Raise when the given browserName config is invalid
    Args:
        msg (str): Human readable string describing the exception.
    """

    def __init__(self, msg):
        self.msg = msg


class BromeBrowserConfigException(Exception):
    """Raise when the given browser config is invalid
    Args:
        msg (str): Human readable string describing the exception.
    """

    def __init__(self, msg):
        self.msg = msg


class ServerBaseException(Exception):
    def get_name(self):
        return self.__repr__().split('(')[0]


class MissingModelValueException(ServerBaseException):
    pass


class WrongEmailOrPasswordException(ServerBaseException):
    pass


class EmailAlreadyExistsException(ServerBaseException):
    pass


class InvalidNameException(ServerBaseException):
    pass


class InvalidPasswordException(ServerBaseException):
    pass


class InvalidEmailException(ServerBaseException):
    pass


class NotAuthorizedException(ServerBaseException):
    pass


class CSRFMismatch(ServerBaseException):
    pass


class InvalidRequestException(ServerBaseException):
    pass


class EmailAlreadyConfirmedException(ServerBaseException):
    pass


class EmailMismatchException(ServerBaseException):
    pass


class TokenExpiredException(ServerBaseException):
    pass


class TokenViolationException(ServerBaseException):
    pass


class TokenAlreadyUsedException(ServerBaseException):
    pass


class TokenInvalidException(ServerBaseException):
    pass


class EmailNotFound(ServerBaseException):
    pass


class ModelImportException(ServerBaseException):
    pass


class InvalidRegistrationTokenException(ServerBaseException):
    pass



import glob
import logging
import os
import os.path

from haikunator import Haikunator
import yaml
import psutil

from brome.core.settings import BROME_CONFIG
from brome.model.test import Test
from brome.model.testbatch import Testbatch
from brome.model.testinstance import Testinstance
from brome.model.testresult import Testresult
from brome.core.utils import (
    DbSessionContext,
    utcnow,
    create_dir_if_doesnt_exist
)


class BaseRunner(object):
    """Base class for the brome runner

    All brome runner inherit from this base class

    Attributes:
        brome (object)
    """

    def __init__(self, brome):
        self.brome = brome
        self.log_file_path = ''

        # Current pid of the runner
        current_pid = os.getpid()

        # Get the activated tests list
        if self.brome.tests:
            if not type(self.brome.tests) == list:
                tests = [self.brome.tests]
            else:
                tests = self.brome.tests

            self.tests = tests
        else:
            self.tests = self.get_activated_tests()

        # Create test batch
        with DbSessionContext(BROME_CONFIG['database']['mongo_database_name']) as session:  # noqa
            self.starting_timestamp = utcnow()

            haikunator = Haikunator()

            self.friendly_name = haikunator.haikunate(
                token_length=0,
                delimiter='.'
            )

            test_batch = Testbatch()
            test_batch.friendly_name = self.friendly_name
            test_batch.starting_timestamp = self.starting_timestamp
            test_batch.pid = current_pid
            if BROME_CONFIG['runner_args']['remote_runner']:
                test_batch.total_tests = len(self.tests) * len(BROME_CONFIG['runner_args']['remote_runner'].split(','))  # noqa
            else:
                test_batch.total_tests = len(self.tests) * len(BROME_CONFIG['runner_args']['localhost_runner'].split(','))  # noqa

            session.save(test_batch, safe=True)

        self.test_batch_id = test_batch.get_uid()
        self.test_batch_friendly_name = test_batch.friendly_name

        # RUNNER LOG DIR
        self.root_test_result_dir = BROME_CONFIG["project"]["test_batch_result_path"]  # noqa

        if self.root_test_result_dir:
            self.runner_dir = os.path.join(
                self.root_test_result_dir,
                "tb_results",
                "tb_%s" % self.friendly_name
            )
            self.relative_runner_dir = os.path.join(
                "tb_results",
                "tb_%s" % self.friendly_name
            )
            create_dir_if_doesnt_exist(self.runner_dir)
        else:
            self.runner_dir = ''

        # LOGGING
        self.configure_logger()

        # SCREENSHOT CACHE
        if BROME_CONFIG['runner']['cache_screenshot']:
            # Dictionary that contains all the screenshot name
            self.screenshot_cache = {}

        with DbSessionContext(BROME_CONFIG['database']['mongo_database_name']) as session:  # noqa
            test_batch = session.query(Testbatch)\
                .filter(Testbatch.mongo_id == self.test_batch_id)\
                .one()

            if self.root_test_result_dir:
                test_batch.root_path = self.root_test_result_dir
                test_batch.log_file_path = self.relative_log_file_path
            else:
                test_batch.root_path = ''
                test_batch.log_file_path = ''

            session.save(test_batch, safe=True)

    def kill_pid(self, pid):
        """Kill process by pid

        Args:
            pid (int)
        """
        try:

            p = psutil.Process(pid)

            p.terminate()

            self.info_log('Killed [pid:%s][name:%s]' % (p.pid, p.name()))
        except psutil.NoSuchProcess:
            self.error_log('No such process: [pid:%s]' % pid)

    def kill(self, procname):
        """Kill by process name

        Args:
            procname (str)
        """
        for proc in psutil.process_iter():
            if proc.name() == procname:
                self.info_log(
                    '[pid:%s][name:%s] killed' %
                    (proc.pid, proc.name())
                )
                proc.kill()

    def get_available_tests(self, search_query=None, test_name=None):
        """Return a list of all the available tests

        This function might call exit(1) if not test were found

        Returns:
            list
        """

        available_tests = []

        script_folder_name = BROME_CONFIG['brome']['script_folder_name']
        script_test_prefix = BROME_CONFIG['brome']['script_test_prefix']

        if search_query:

            tests_path = os.path.join(
                BROME_CONFIG['project']['absolute_path'],
                script_folder_name,
                '%s%s.py' % (script_test_prefix, search_query)
            )
            tests = sorted(glob.glob(tests_path))

            for test in tests:
                module_test = test.split(os.sep)[-1][:-3]
                available_tests.append(
                    __import__(
                        '%s.%s' %
                        (script_folder_name, module_test),
                        fromlist=[''])
                )

        elif test_name:
            if test_name.endswith('.py'):
                test_name = test_name[:-3]

            try:
                available_tests.append(
                    __import__(
                        '%s.%s' %
                        (script_folder_name, test_name),
                        fromlist=[''])
                    )
            except ImportError:
                pass

        if not len(available_tests):
            query = search_query if search_query else test_name
            print("No script found with the provided query: %s" % query)
            exit(1)

        return available_tests

    def get_activated_tests(self):
        """Return a list of all the activated tests

        The behavior of this function is determined by
        the received CLI arguments

        Returns:
            list
        """
        tests = []

        # test file
        if BROME_CONFIG['runner_args']['test_file']:
            test_file_path = BROME_CONFIG['runner_args']['test_file']
            with open(test_file_path, 'r') as f:
                test_list = yaml.load(f)

            for test in test_list:
                tests.append(self.get_available_tests(test)[0])

        elif BROME_CONFIG['runner_args']['test_name']:
            tests = self.get_available_tests(
                test_name=BROME_CONFIG['runner_args']['test_name']
            )
        else:
            test_search_query = BROME_CONFIG['runner_args']['test_search_query']  # noqa

            # by index or slice e.g.: [0:12], [:], [0], [-1]
            if test_search_query.find('[') != -1:
                exec(
                    'tests = self.get_available_tests("*")%s' %
                    test_search_query
                )

            # by name
            else:
                tests = self.get_available_tests(test_search_query)

        return tests

    def configure_logger(self):
        """Configure the test batch runner logger
        """

        logger_name = 'brome_runner'

        self.logger = logging.getLogger(logger_name)

        format_ = BROME_CONFIG['logger_runner']['format']

        # Stream logger
        if BROME_CONFIG['logger_runner']['streamlogger']:
            sh = logging.StreamHandler()
            stream_formatter = logging.Formatter(format_)
            sh.setFormatter(stream_formatter)
            self.logger.addHandler(sh)

        # File logger
        if BROME_CONFIG['logger_runner']['filelogger'] and \
                self.runner_dir:

            self.log_file_path = os.path.join(
                    self.runner_dir,
                    '%s.log' % logger_name
            )
            self.relative_log_file_path = os.path.join(
                    self.relative_runner_dir,
                    '%s.log' % logger_name
            )

            fh = logging.FileHandler(
                self.log_file_path
            )
            file_formatter = logging.Formatter(format_)
            fh.setFormatter(file_formatter)
            self.logger.addHandler(fh)

        self.logger.setLevel(
            getattr(
                logging,
                BROME_CONFIG['logger_runner']['level']
                )
            )

    def print_test_summary(self, executed_tests):
        """Print test summary

        When the test batch is finished a test summary will be printed

        Args:
            executed_tests (list)
        """

        separator = '---------------------'

        with DbSessionContext(BROME_CONFIG['database']['mongo_database_name']) as session:  # noqa
            test_batch = session.query(Testbatch).filter(Testbatch.mongo_id == self.test_batch_id).one()  # noqa

            # TITLE
            self.info_log('******* TEST BATCH SUMMARY ********')

            # TOTAL NUMBER OF EXECUTED TESTS
            base_query = session.query(Testresult).filter(Testresult.test_batch_id == self.test_batch_id)  # noqa
            total_test = base_query.count()
            total_test_successful = base_query.filter(Testresult.result == True).count()  # noqa
            base_query = session.query(Testresult).filter(Testresult.test_batch_id == self.test_batch_id)  # noqa
            total_test_failed = base_query.filter(Testresult.result == False).count()  # noqa
            self.info_log(
                'Total_test: %s; Total_test_successful: %s; Total_test_failed: %s' %  # noqa
                (total_test, total_test_successful, total_test_failed)
            )

            # EXECUTION TIME
            self.info_log(
                "Total execution time: %s" %
                (test_batch.ending_timestamp - test_batch.starting_timestamp)
            )

            # SEPARATOR
            self.info_log(separator)

            self.info_log('Failed tests:')

            # FAILED TESTS
            failed_test_list = []
            test_results = session.query(Testresult)\
                .filter(Testresult.result == False)\
                .filter(Testresult.test_batch_id == self.test_batch_id).all()  # noqa
            for test_result in test_results:
                if test_result.title not in failed_test_list:
                    failed_test_list.append(test_result.title)
                    query = session.query(Test)\
                        .filter(Test.mongo_id == test_result.test_id)
                    if query.count():
                        test = query.one()
                        self.info_log(
                            "[%s] %s" %
                            (test.test_id, test.name)
                        )
                    else:
                        self.info_log(
                            "[noid] %s" %
                            (test_result.title)
                        )

            if not failed_test_list:
                self.info_log('No test failed!')

            # SEPARATOR
            self.info_log(separator)

            # TEST INSTANCE REPORT
            for test in executed_tests:
                # TITLE
                self.info_log(
                    '%s %s' %
                    (test._name, test.pdriver.get_id())
                )

                test_instance = session.query(Testinstance)\
                    .filter(Testinstance.mongo_id == test._test_instance_id)\
                    .one()

                # TEST EXECUTION TIME
                try:
                    self.info_log(
                        "Test execution time: %s" %
                        (test_instance.ending_timestamp - test_instance.starting_timestamp)  # noqa
                    )
                except TypeError:
                    self.info_log("Test execution time exception")

                # TEST INSTANCE SUMMARY
                results = test.get_test_result_summary()
                for result in results:
                    self.info_log(result)

                # CRASH REPORT
                if test._crash_error:
                    self.info_log(test._crash_error)
                else:
                    self.info_log('No crash!')

                # SEPARATOR
                self.info_log(separator)

            # END
            self.info_log('Finished')

    def get_logger_dict(self):
        return {'batchid': self.test_batch_friendly_name}

    def debug_log(self, msg):
        self.logger.debug(
            "[debug]%s" %
            msg, extra=self.get_logger_dict()
        )

    def info_log(self, msg):
        self.logger.info(
            "%s" %
            msg, extra=self.get_logger_dict()
        )

    def warning_log(self, msg):
        self.logger.warning(
            "[warning]%s" %
            msg, extra=self.get_logger_dict()
        )

    def error_log(self, msg):
        self.logger.error(
            "[error]%s" %
            msg, extra=self.get_logger_dict()
        )

    def critical_log(self, msg):
        self.logger.critical(
            "[critical]%s" %
            msg, extra=self.get_logger_dict()
        )



from mongoalchemy.session import Session
from dateutil import tz

from brome.core.settings import BROME_CONFIG

async def db_handler(app, handler):
    async def middleware(request):
        if request.path.startswith('/api/'):
            request.db_session = Session.connect(
                BROME_CONFIG['database'].get("mongo_database_name"),
                timezone=tz.gettz('UTC')
            )

        response = await handler(request)
        return response

    return middleware



from mongoalchemy.fields import (
    StringField,
    AnythingField,
    ObjectIdField,
    EnumField,
    DictField
)

from brome.model.basemodel import BaseModel
from brome.core import exceptions


class Testscreenshot(BaseModel):
    browser_capabilities = DictField(AnythingField())
    browser_id = StringField()
    file_path = StringField()
    root_path = StringField()
    location = EnumField(StringField(), 's3', 'local_file_system')
    extra_data = DictField(StringField(), default=dict())
    title = StringField()

    test_instance_id = ObjectIdField()
    test_batch_id = ObjectIdField()

    def __repr__(self):
        try:
            return "Testscreenshot <uid: {self.mongo_id}><test_batch_id: {self.test_batch_id}>".format(  # noqa
                self=self
            )
        except AttributeError:
            return "Testscreenshot uninitialized"

    async def sanitize_data(self, context):
        return []

    async def serialize(self, context):
        data = {}
        data['uid'] = self.get_uid()
        data['root_path'] = self.root_path
        data['file_path'] = self.file_path
        data['title'] = self.title
        data['browser_capabilities'] = self.browser_capabilities
        return data

    async def method_autorized(self, context):
        method = context.get('method')
        if method in ['create', 'update', 'delete']:
            return False
        else:
            return True

    async def validate_and_save(self, context):
        data = context.get('data')
        db_session = context.get('db_session')

        is_new = await self.is_new()

        # BROWSER ID
        browser_id = data.get('browser_id')
        if browser_id:
            self.browser_id = browser_id
        else:
            if is_new:
                raise exceptions.MissingModelValueException('browser_id')

        # BROWSER CAPABILITIES
        browser_capabilities = data.get('browser_capabilities')
        if browser_capabilities:
            self.browser_capabilities = browser_capabilities
        else:
            if is_new:
                raise exceptions.MissingModelValueException(
                    'browser_capabilities'
                )

        # FILE PATH
        file_path = data.get('file_path')
        if file_path:
            self.file_path = file_path
        else:
            if is_new:
                raise exceptions.MissingModelValueException('file_path')

        # ROOT PATH
        root_path = data.get('root_path')
        if root_path:
            self.root_path = root_path
        else:
            if is_new:
                raise exceptions.MissingModelValueException('root_path')

        # LOCATION
        location = data.get('location')
        if location:
            self.location = location
        else:
            if is_new:
                self.location = 'local_file_system'

        # EXTRA_DATA
        extra_data = data.get('extra_data')
        if extra_data:
            self.extra_data = extra_data

        # TITLE
        title = data.get('title')
        if title:
            self.title = title
        else:
            if is_new:
                raise exceptions.MissingModelValueException('title')

        # TEST_INSTANCE_ID
        test_instance_id = data.get('test_instance_id')
        if test_instance_id:
            self.test_instance_id = test_instance_id
        else:
            if is_new:
                raise exceptions.MissingModelValueException('test_instance_id')

        # TEST_BATCH_ID
        test_batch_id = data.get('test_batch_id')
        if test_batch_id:
            self.test_batch_id = test_batch_id
        else:
            if is_new:
                raise exceptions.MissingModelValueException('test_batch_id')

        db_session.save(self, safe=True)



import importlib
import traceback
import logging

from aiohttp_session import get_session
from aiohttp import web

from brome.core import exceptions
from brome.webserver.server.server_decorator import (
    require,
    exception_handler,
    csrf_protected
)
from brome.webserver.server.auth import get_user_from_session

logger = logging.getLogger('bromewebserver')


class CRUD(web.View):

    action_list = [
        'create',
        'read',
        'update',
        'delete'
    ]

    async def get_json_data(self):
        try:
            req_data = await self.request.json()
            actions = req_data['actions']

            # Convert actions to list if necessary
            if not type(actions) == list:
                actions = [actions]

            logger.debug('actions = {actions}'.format(
                actions=actions
            ))

            return actions
        except:
            raise exceptions.InvalidRequestException(
                'No json send or missing actions parameters'
            )

    def import_model(self, model):
        try:
            m = importlib.import_module(
                'brome.model.{model}'.format(model=model)
            )
            return getattr(m, model.title())
        except ImportError:
            raise exceptions.ModelImportException(
                '{model} not found'.format(model=model)
            )

    def trim_response_data(self, success, response_data, error):
        if len(response_data) == 1:
            return response_data[0]
        else:
            data = {'success': success, 'results': response_data}
            if error:
                data['error'] = error
            return data

    @exception_handler()
    @csrf_protected()
    @require('login')
    async def post(self):
        actions = await self.get_json_data()

        session = await get_session(self.request)
        author = get_user_from_session(session, self.request.db_session)

        read_context = {
            'author': author,
            'db_session': self.request.db_session,
            'ws_session': session,
            'method': 'read',
            'queue': self.request.app.queue,
        }
        action_context = {
            'author': author,
            'db_session': self.request.db_session,
            'ws_session': session,
            'queue': self.request.app.queue
        }

        success = True
        response_data = []
        error = []
        for index, action in enumerate(actions):
            try:
                # RESPONSE
                response_data.append(dict())
                response_data[index]['success'] = True
                response_data[index]['results'] = []

                # ACTION
                action_name = action.get('action')
                if not action_name:
                    raise exceptions.InvalidRequestException(
                        'Missing action in actions[{index}]'
                        .format(
                            index=index
                        )
                    )
                elif action_name not in self.action_list:
                    raise exceptions.InvalidRequestException(
                        'Invalid action name: "{action_name}"'
                        .format(
                            action_name=action_name
                        )
                    )

                action_context['method'] = action_name

                # IMPORT MODEL
                model_name = action.get('model')
                if not model_name:
                    raise exceptions.InvalidRequestException(
                        'Missing model in action'
                    )

                model_class = self.import_model(model_name)

                # CREATE
                if action_name == 'create':
                    results = [model_class()]
                    response_data[index]['total'] = 1

                # QUERY
                else:
                    # READ SPECIFIC RECORD
                    uid = action.get('uid')
                    if uid:
                        base_query = self.request.db_session\
                            .query(model_class)\
                            .filter(model_class.mongo_id == uid)

                    # BATCH
                    elif action.get('uids'):
                        uids = action.get('uids')
                        base_query = self.request.db_session.query(model_class)
                        response_data[index]['total'] = len(uids)
                        base_query = base_query.in_('mongo_id', *uids)
                        results = base_query.all()

                    else:
                        filters = action.get('filters')
                        limit = action.get('limit')
                        skip = action.get('skip')
                        descending = action.get('descending')
                        ascending = action.get('ascending')

                        base_query = self.request.db_session.query(model_class)

                        if limit:
                            base_query = base_query.limit(limit)

                        if skip:
                            base_query = base_query.skip(skip)

                        if descending:
                            base_query = base_query.descending(descending)

                        if ascending:
                            base_query = base_query.ascending(ascending)

                        if filters:
                            if 'uid' in filters:
                                filters['mongo_id'] = filters['uid']
                                del filters['uid']

                            base_query = base_query.filter_by(**filters)

                    response_data[index]['total'] = base_query.count()
                    results = base_query.all()

                # PROCESSING RESULTS
                for result in results:
                    # AUTHORIZATION CHECK
                    logger.debug(
                        'action_context = {action_context}'
                        .format(
                            action_context=action_context
                        )
                    )
                    if not await result.method_autorized(action_context):
                        raise exceptions.NotAuthorizedException(
                            '{author} not authorized to {action_name} {result}'
                            .format(
                                author=author,
                                action_name=action_name,
                                result=result
                            )
                        )

                    # APPLY ACTION
                    # CREATE & UPDATE
                    if action_name in ['create', 'update']:
                        data = action.get('data')
                        if not data:
                            raise exceptions.InvalidRequestException(
                                'Missing data in action'
                            )

                        action_context['data'] = data
                        sane_data = await result.sanitize_data(
                            action_context
                        )
                        action_context['data'] = sane_data
                        # BEFORE HOOK
                        await getattr(
                            result,
                            'before_{action_name}'
                            .format(
                                action_name=action_name
                            )
                        )(action_context)

                        await result.validate_and_save(action_context)

                        # AFTER HOOK
                        await getattr(
                            result,
                            'after_{action_name}'
                            .format(
                                action_name=action_name
                            )
                        )(action_context)


                    # DELETE
                    elif action_name == 'delete':
                        await result.before_delete(action_context)
                        self.request.db_session.remove(result, safe=True)
                        await result.after_delete(action_context)

                    if not action.get('total_only', False) \
                            and not action_name == 'delete':

                        # READ
                        # NOTE the authorization check has already
                        # been performed for the read
                        if not action_name == 'read':
                            logger.debug(
                                'read_context = {read_context}'
                                .format(
                                    read_context=read_context
                                )
                            )
                            if not await result.method_autorized(read_context):
                                raise exceptions.NotAuthorizedException(
                                    '{author} not authorized to {action_name} {result}'  # noqa
                                    .format(
                                        author=author,
                                        action_name=read_context.get('method'),
                                        result=result
                                    )
                                )
                        response_data[index]['results'].append(
                            await result.serialize(read_context)
                        )
            except Exception as e:
                success = False
                tb = traceback.format_exc()
                logger.error(
                    'Request HandledException<{exception}>'
                    .format(exception=str(tb))
                )
                if isinstance(e, exceptions.ServerBaseException):
                    error_msg = e.get_name()
                else:
                    error_msg = 'ServerSideError'

                response_data[index] = {
                    'success': False,
                    'error': error_msg
                }
                error.append(error_msg)

        # RESPONSE
        trimmed_response_data = self.trim_response_data(
            success,
            response_data,
            error
        )
        return web.json_response(trimmed_response_data)
